import os
import subprocess
import pprint
import glob
from shutil import copy2
import platform
import pytest
import rsgislib.tools.filetools

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
RASTERGIS_DATA_DIR = os.path.join(DATA_DIR, "rastergis")
IMGUTILS_DATA_DIR = os.path.join(DATA_DIR, "imageutils")

# TODO Need to figure out what is happening on windows but get it building first!
on_windows = platform.system() == "Windows"

JINJA2_NOT_AVAIL = False
try:
    import jinja2
except ImportError:
    JINJA2_NOT_AVAIL = True

gdal2tiles_cmd_avail = rsgislib.tools.filetools.is_cmd_tool_avail(
    "gdal2tiles.py", test_call_cmd=["gdal2tiles.py", "-h"]
)
gdal_translate_cmd_avail = rsgislib.tools.filetools.is_cmd_tool_avail(
    "gdal_translate", test_call_cmd=["gdal_translate", "-h"]
)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgis_config_version():
    rtn_info = subprocess.run(
        ["rsgis-config", "--version"], capture_output=True, text=True, check=True
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("5" in rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgis_config_prefix():
    rtn_info = subprocess.run(
        ["rsgis-config", "--prefix"], capture_output=True, text=True, check=True
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgis_config_libs():
    rtn_info = subprocess.run(
        ["rsgis-config", "--libs"], capture_output=True, text=True, check=True
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgis_config_cflags():
    rtn_info = subprocess.run(
        ["rsgis-config", "--cflags"], capture_output=True, text=True, check=True
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgis_config_ldflags():
    rtn_info = subprocess.run(
        ["rsgis-config", "--ldflags"], capture_output=True, text=True, check=True
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgis_config_libdir():
    rtn_info = subprocess.run(
        ["rsgis-config", "--libdir"], capture_output=True, text=True, check=True
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgis_config_includes():
    rtn_info = subprocess.run(
        ["rsgis-config", "--includes"], capture_output=True, text=True, check=True
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgis_config_fail():
    rtn_info = subprocess.run(
        ["rsgis-config", "--hello"], capture_output=True, text=True, check=True
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" == rtn_info.stdout)


@pytest.mark.skipif(
    (JINJA2_NOT_AVAIL or on_windows or (not gdal_translate_cmd_avail)),
    reason="Either jinja2 not available or command not tested on Windows, "
    "or gdal_translate is not available",
)
def test_rsgisapplycmd_run(tmp_path):
    in_dir = os.path.join(tmp_path, "in")
    os.mkdir(in_dir)
    out_dir = os.path.join(tmp_path, "out")
    os.mkdir(out_dir)

    img1_in = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    img1_out = os.path.join(in_dir, "sen2_20210527_aber_vldmsk.kea")
    copy2(img1_in, img1_out)
    img2_in = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    img2_out = os.path.join(in_dir, "sen2_20210527_aber_subset_vldmsk.kea")
    copy2(img2_in, img2_out)

    rtn_info = subprocess.run(
        [
            "rsgisapplycmd.py",
            "--indir",
            in_dir,
            "--outdir",
            out_dir,
            "--inext",
            "kea",
            "--outext",
            "tif",
            "--nameapp",
            "test",
            "--cmd",
            "gdal_translate -of GTIFF {{ input }} {{ output }}",
            "--ncores",
            "2",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and (
        len(glob.glob(os.path.join(out_dir, "*.tif"))) == 2
    )


@pytest.mark.skipif(
    (JINJA2_NOT_AVAIL or on_windows or (not gdal_translate_cmd_avail)),
    reason="Either jinja2 not available or command not tested on "
    "Windows or gdal_translate is not available",
)
def test_rsgisapplycmd_print(tmp_path):
    in_dir = os.path.join(tmp_path, "in")
    os.mkdir(in_dir)
    out_dir = os.path.join(tmp_path, "out")
    os.mkdir(out_dir)

    img1_in = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    img1_out = os.path.join(in_dir, "sen2_20210527_aber_vldmsk.kea")
    copy2(img1_in, img1_out)
    img2_in = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    img2_out = os.path.join(in_dir, "sen2_20210527_aber_subset_vldmsk.kea")
    copy2(img2_in, img2_out)

    rtn_info = subprocess.run(
        [
            "rsgisapplycmd.py",
            "--indir",
            in_dir,
            "--outdir",
            out_dir,
            "--inext",
            "kea",
            "--outext",
            "tif",
            "--nameapp",
            "test",
            "--cmd",
            "gdal_translate -of GTIFF {{ input }} {{ output }}",
            "--printcmds",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("gdal_translate" in rtn_info.stdout)


@pytest.mark.skipif(
    ((not gdal_translate_cmd_avail) or on_windows),
    reason="Command not tested on Windows or gdal_translate is not available",
)
def test_rsgisbatchconvert2tif(tmp_path):
    in_dir = os.path.join(tmp_path, "in")
    os.mkdir(in_dir)
    out_dir = os.path.join(tmp_path, "out")
    os.mkdir(out_dir)

    img1_in = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    img1_out = os.path.join(in_dir, "sen2_20210527_aber_vldmsk.kea")
    copy2(img1_in, img1_out)
    img2_in = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    img2_out = os.path.join(in_dir, "sen2_20210527_aber_subset_vldmsk.kea")
    copy2(img2_in, img2_out)

    in_srch = os.path.join(in_dir, "*.kea")
    rtn_info = subprocess.run(
        ["rsgisbatchconvert2tif.py", "-i", in_srch, "-o", out_dir],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and (
        len(glob.glob(os.path.join(out_dir, "*.tif"))) == 2
    )


@pytest.mark.skipif(
    ((not gdal_translate_cmd_avail) or on_windows),
    reason="Command not tested on Windows or gdal_translate is not available",
)
def test_rsgisbatchconvert2tif_chk(tmp_path):
    in_dir = os.path.join(tmp_path, "in")
    os.mkdir(in_dir)
    out_dir = os.path.join(tmp_path, "out")
    os.mkdir(out_dir)

    img1_in = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    img1_out = os.path.join(in_dir, "sen2_20210527_aber_vldmsk.kea")
    copy2(img1_in, img1_out)
    img2_in = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    img2_out = os.path.join(in_dir, "sen2_20210527_aber_subset_vldmsk.kea")
    copy2(img2_in, img2_out)

    in_srch = os.path.join(in_dir, "*.kea")
    rtn_info = subprocess.run(
        ["rsgisbatchconvert2tif.py", "-i", in_srch, "-o", out_dir, "--chkimgs"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and (
        len(glob.glob(os.path.join(out_dir, "*.tif"))) == 2
    )


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisbuildimglut(tmp_path):
    in_dir = os.path.join(tmp_path, "in")
    os.mkdir(in_dir)

    img1_in = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    img1_out = os.path.join(in_dir, "sen2_20210527_aber_vldmsk.kea")
    copy2(img1_in, img1_out)
    img2_in = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    img2_out = os.path.join(in_dir, "sen2_20210527_aber_subset_vldmsk.kea")
    copy2(img2_in, img2_out)

    in_srch = os.path.join(in_dir, "*.kea")
    out_lut_file = os.path.join(tmp_path, "lut.gpkg")
    rtn_info = subprocess.run(
        [
            "rsgisbuildimglut.py",
            "-i",
            in_srch,
            "-o",
            out_lut_file,
            "--veclyr",
            "test_lut",
            "--vecformat",
            "GPKG",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and os.path.exists(out_lut_file)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisbuildimglut_wgs84(tmp_path):
    in_dir = os.path.join(tmp_path, "in")
    os.mkdir(in_dir)

    img1_in = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    img1_out = os.path.join(in_dir, "sen2_20210527_aber_vldmsk.kea")
    copy2(img1_in, img1_out)
    img2_in = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    img2_out = os.path.join(in_dir, "sen2_20210527_aber_subset_vldmsk.kea")
    copy2(img2_in, img2_out)

    in_srch = os.path.join(in_dir, "*.kea")
    out_lut_file = os.path.join(tmp_path, "lut.gpkg")
    rtn_info = subprocess.run(
        [
            "rsgisbuildimglut.py",
            "-i",
            in_srch,
            "-o",
            out_lut_file,
            "--veclyr",
            "test_lut",
            "--vecformat",
            "GPKG",
            "--wgs84",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and os.path.exists(out_lut_file)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgiscalcimgstats_cont(tmp_path):
    img_ref = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    img = os.path.join(tmp_path, "sen2_20210527_aber.kea")
    copy2(img_ref, img)

    rtn_info = subprocess.run(
        ["rsgiscalcimgstats.py", "-i", img, "-n", "0.0"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert rtn_info.returncode == 0


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgiscalcimgstats_therm(tmp_path):
    img_ref = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    img = os.path.join(tmp_path, "sen2_20210527_aber_vldmsk.kea")
    copy2(img_ref, img)

    rtn_info = subprocess.run(
        ["rsgiscalcimgstats.py", "-i", img, "-n", "0.0", "-t"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert rtn_info.returncode == 0


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgiscalcimgstats_therm_tif(tmp_path):
    img_ref = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.tif")
    img = os.path.join(tmp_path, "sen2_20210527_aber_vldmsk.tif")
    copy2(img_ref, img)

    rtn_info = subprocess.run(
        ["rsgiscalcimgstats.py", "-i", img, "-n", "0.0", "-t"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert rtn_info.returncode == 0


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgischkgdalfile_sgl_img(tmp_path):
    img_ref = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    img = os.path.join(tmp_path, "sen2_20210527_aber_vldmsk.kea")
    copy2(img_ref, img)

    rtn_info = subprocess.run(
        [
            "rsgischkgdalfile.py",
            "-i",
            img,
            "--nbands",
            "1",
            "--epsg",
            "27700",
            "--chkproj",
            "--readimg",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert rtn_info.returncode == 0


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgischkgdalfile_multi_img(tmp_path):
    img_ref = os.path.join(DATA_DIR, "sen2_20210527_aber_vldmsk.kea")
    img = os.path.join(tmp_path, "sen2_20210527_aber_vldmsk.kea")
    copy2(img_ref, img)

    img_ref = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    img = os.path.join(tmp_path, "sen2_20210527_aber_subset_vldmsk.kea")
    copy2(img_ref, img)

    in_srch = os.path.join(tmp_path, "*.kea")
    rtn_info = subprocess.run(
        [
            "rsgischkgdalfile.py",
            "-i",
            in_srch,
            "--nbands",
            "1",
            "--epsg",
            "27700",
            "--chkproj",
            "--readimg",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert rtn_info.returncode == 0


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgischkgdalfile_sgl_vec(tmp_path):
    vec_ref = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec = os.path.join(tmp_path, "aber_osgb_multi_polys.geojson")
    copy2(vec_ref, vec)

    rtn_info = subprocess.run(
        ["rsgischkgdalfile.py", "-i", vec, "--vec", "--epsg", "27700", "--chkproj"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert rtn_info.returncode == 0


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgischkgdalfile_multi_vec(tmp_path):
    vec_ref = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec = os.path.join(tmp_path, "aber_osgb_multi_polys.geojson")
    copy2(vec_ref, vec)

    vec_ref = os.path.join(DATA_DIR, "aber_osgb_single_poly_hole.geojson")
    vec = os.path.join(tmp_path, "aber_osgb_single_poly_hole.geojson")
    copy2(vec_ref, vec)

    in_srch = os.path.join(tmp_path, "*.geojson")
    rtn_info = subprocess.run(
        ["rsgischkgdalfile.py", "-i", in_srch, "--vec", "--epsg", "27700", "--chkproj"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert rtn_info.returncode == 0


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgiscopybandnames(tmp_path):
    img_ref = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    img = os.path.join(tmp_path, "sen2_20210527_aber.kea")
    copy2(img_ref, img)

    img_bandnames_ref = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")

    rtn_info = subprocess.run(
        ["rsgiscopybandnames.py", "-r", img_bandnames_ref, "-o", img],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert rtn_info.returncode == 0


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_sha1():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "SHA1"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_sha224():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "SHA224"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_sha256():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "SHA256"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_sha384():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "SHA384"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_sha512():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "SHA512"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_md5():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "MD5"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_blake2b():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "Blake2B"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_black2s():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "Blake2S"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_sha3_224():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "SHA3_224"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_sha3_256():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "SHA3_256"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_sha3_234():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "SHA3_384"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilehash_sha3_512():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisfilehash.py", "-i", img, "-o", "SHA3_512"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilelut_copy(tmp_path):
    import rsgislib.imageutils.imagelut
    import glob

    input_imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    vec_lut_file = os.path.join(tmp_path, "test_lut.gpkg")
    vec_lut_lyr = "test_lut"
    rsgislib.imageutils.imagelut.create_img_extent_lut(
        input_imgs,
        vec_lut_file,
        vec_lut_lyr,
        "GPKG",
        ignore_none_imgs=False,
        out_proj_wgs84=False,
        overwrite_lut_file=False,
    )

    roi_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    roi_lyr = "aber_osgb_single_poly"

    rtn_info = subprocess.run(
        [
            "rsgisfilelut.py",
            "-i",
            vec_lut_file,
            "--lutlyr",
            vec_lut_lyr,
            "--roifile",
            roi_file,
            "--roilyr",
            roi_lyr,
            "--dest",
            tmp_path,
            "--copy",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisfilelut_targz(tmp_path):
    import rsgislib.imageutils.imagelut
    import glob

    input_imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    vec_lut_file = os.path.join(tmp_path, "test_lut.gpkg")
    vec_lut_lyr = "test_lut"
    rsgislib.imageutils.imagelut.create_img_extent_lut(
        input_imgs,
        vec_lut_file,
        vec_lut_lyr,
        "GPKG",
        ignore_none_imgs=False,
        out_proj_wgs84=False,
        overwrite_lut_file=False,
    )

    roi_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    roi_lyr = "aber_osgb_single_poly"

    out_file = os.path.join(tmp_path, "out.tar.gz")

    rtn_info = subprocess.run(
        [
            "rsgisfilelut.py",
            "-i",
            vec_lut_file,
            "--lutlyr",
            vec_lut_lyr,
            "--roifile",
            roi_file,
            "--roilyr",
            roi_lyr,
            "--dest",
            out_file,
            "--targz",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisimg2kmz(tmp_path):
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_file = os.path.join(tmp_path, "sen2_20210527_aber.kmz")

    rtn_info = subprocess.run(
        ["rsgisimg2kmz.py", "-i", img, "-o", out_file, "-b", "3,2,1"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and (os.path.exists(out_file))


@pytest.mark.skipif(
    (not gdal2tiles_cmd_avail) or on_windows,
    reason="Command not tested on Windows or gdal2tiles.py command not available.",
)
def test_rsgisimg2webtiles(tmp_path):
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisimg2webtiles.py", "-i", img, "-o", tmp_path, "-b", "3,2,1"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert rtn_info.returncode == 0


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisimginfo():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisimginfo.py", "-i", img], capture_output=True, text=True, check=True
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisimginfo_rat():
    img = os.path.join(RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_attref.kea")

    rtn_info = subprocess.run(
        ["rsgisimginfo.py", "-i", img], capture_output=True, text=True, check=True
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisproj_img_epsg():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisproj.py", "--image", img, "-o", "EPSG"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisproj_img_wkt():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisproj.py", "--image", img, "-o", "WKT"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisproj_img_wktpretty():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisproj.py", "--image", img, "-o", "WKTPretty"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisproj_img_proj4():
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    rtn_info = subprocess.run(
        ["rsgisproj.py", "--image", img, "-o", "PROJ4"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisproj_utm_36n_epsg():
    rtn_info = subprocess.run(
        ["rsgisproj.py", "--utm", "36N", "-o", "EPSG"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisproj_utm_01n_epsg():
    rtn_info = subprocess.run(
        ["rsgisproj.py", "--utm", "01N", "-o", "EPSG"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisproj_utm_55s_epsg():
    rtn_info = subprocess.run(
        ["rsgisproj.py", "--utm", "55S", "-o", "EPSG"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisproj_wkt_epsg():
    file_path = os.path.join(IMGUTILS_DATA_DIR, "utm30n.wkt")

    rtn_info = subprocess.run(
        ["rsgisproj.py", "--wktfile", file_path, "-o", "EPSG"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisproj_epsg_wkt():
    rtn_info = subprocess.run(
        ["rsgisproj.py", "--epsg", "27700", "-o", "WKT"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(
    ((not gdal_translate_cmd_avail) or on_windows),
    reason="Command not tested on Windows or gdal_translate is not available",
)
def test_rsgistranslate2tif(tmp_path):
    img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_img = os.path.join(tmp_path, "out_img.tif")
    rtn_info = subprocess.run(
        ["rsgistranslate2tif.py", "-i", img, "-o", out_img],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and os.path.exists(out_img)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisvectools_lyrs():
    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    rtn_info = subprocess.run(
        ["rsgisvectools.py", "--vecfile", vec_file, "--lyrs"],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisvectools_cols():
    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    rtn_info = subprocess.run(
        [
            "rsgisvectools.py",
            "--vecfile",
            vec_file,
            "--veclyr",
            "aber_osgb_multi_polys",
            "--cols",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisvectools_nfeats():
    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    rtn_info = subprocess.run(
        [
            "rsgisvectools.py",
            "--vecfile",
            vec_file,
            "--veclyr",
            "aber_osgb_multi_polys",
            "--nfeats",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)


@pytest.mark.skipif(on_windows, reason="Command not tested on Windows")
def test_rsgisvectools_proj():
    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    rtn_info = subprocess.run(
        [
            "rsgisvectools.py",
            "--vecfile",
            vec_file,
            "--veclyr",
            "aber_osgb_multi_polys",
            "--proj",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    pprint.pprint(rtn_info)
    assert (rtn_info.returncode == 0) and ("" != rtn_info.stdout)



#!/usr/bin/env python
"""
Tools for accessing (searching and downloading) datasets from the NASA
EOSDIS Common Metadata Repository API (https://cmr.earthdata.nasa.gov/search/)

"""

import datetime
from typing import Union, List, Dict
import os

import requests
import requests.utils

import rsgislib
import rsgislib.tools.utils
import rsgislib.tools.httptools

# CMR_OPS = "https://cmr.earthdata.nasa.gov/search/"
# CMR_UAT = "https://cmr.uat.earthdata.nasa.gov/search/"
# CMR_SIT = "https://cmr.sit.earthdata.nasa.gov/search/"

CMR_COLLECTS_URL = "https://cmr.earthdata.nasa.gov/search/collections.json"
CMR_GRANULES_URL = "https://cmr.earthdata.nasa.gov/search/granules.json"


class SessionWithHeaderNASARedirection(requests.Session):
    AUTH_HOST = "urs.earthdata.nasa.gov"

    def __init__(self, username, password):
        super().__init__()
        self.auth = (username, password)

    def rebuild_auth(self, prepared_request, response):
        """
        Overrides from the library to keep headers when redirected to or from
        the NASA auth host.

        :param prepared_request:
        :param response:
        :return:
        """
        headers = prepared_request.headers
        url = prepared_request.url

        if "Authorization" in headers:
            original_parsed = requests.utils.urlparse(response.request.url)
            redirect_parsed = requests.utils.urlparse(url)
            if (
                (original_parsed.hostname != redirect_parsed.hostname)
                and redirect_parsed.hostname != self.AUTH_HOST
                and original_parsed.hostname != self.AUTH_HOST
            ):
                del headers["Authorization"]

        return


def _check_cmr_response(data: Dict) -> Union[str, List, Dict]:
    """
    A function which checks the response for an error (producing an exception)
    and extracts the text output and returns it.

    :param data:
    :return:

    """

    if not rsgislib.tools.utils.dict_struct_does_path_exist(data, ["feed", "entry"]):
        print(data)
        raise rsgislib.RSGISPyException("Data structure is not as expected - check.")

    return data["feed"]["entry"]


def get_prods_info(prod_short_name: str) -> List[Dict]:
    """
    A function which returns information for a product available from the CMR.

    Available products can be found here:
    https://earthdata.nasa.gov/eosdis/science-system-description/eosdis-standard-products

    :param prod_short_name: The name of the product you are interested in.
    :return: A list of products (probably different versions).

    """
    prod_srch_params = {"short_name": prod_short_name}
    header_info = {"Accept": "application/json"}

    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        CMR_COLLECTS_URL,
        data=prod_srch_params,
        convert_to_json=False,
        header_data=header_info,
    )
    prod_lst = _check_cmr_response(rtnd_data)

    return prod_lst


def check_prod_version_avail(prod_short_name: str, version: str) -> bool:
    """
    A function which checks if a version is available.

    :param prod_short_name: the product short name for the product of interest.
    :param version: the version of the product to be retrieved.
    :return: Boolean specifying whether the version is available.

    """
    prod_lst = get_prods_info(prod_short_name)
    found_version = False
    for prod in prod_lst:
        if prod["version_id"] == version:
            found_version = True
            break
    return found_version


def get_max_prod_version(prod_short_name: str) -> str:
    """
    A function which attempts to find the highest (latest) version for
    a product.

    :param prod_short_name: the product short name for the product of interest.
    :return: string representation of the highest version.

    """
    prod_lst = get_prods_info(prod_short_name)
    first = True
    max_version_int = 0
    max_version = ""
    for prod in prod_lst:
        version_int = int(prod["version_id"])
        if first:
            max_version_int = version_int
            max_version = prod["version_id"]
            first = False
        elif version_int > max_version_int:
            max_version_int = version_int
            max_version = prod["version_id"]
    return max_version


def find_granules(
    prod_short_name: str,
    version: str,
    only_dnwld: bool = True,
    bbox: List[float] = None,
    pt: List[float] = None,
    start_date: datetime.datetime = None,
    end_date: datetime.datetime = None,
    cloud_min: int = 0,
    cloud_max: int = None,
    sort_date: bool = True,
    sort_desc: bool = True,
    page_size: int = 100,
    page_num: int = 1,
    other_params: Dict[str, str] = None,
) -> List[Dict]:
    """
    A function which will find granules from the CMR system for the product of
    interest using the search parameters provided.

    https://cmr.earthdata.nasa.gov/search/site/docs/search/api.html#granule-search-by-parameters

    :param prod_short_name: the product short name for the product of interest.
    :param version: the version of the product to be retrieved.
    :param only_dnwld: If true (default)
    :param bbox: (MinX, MaxX, MinY, MaxY)
    :param pt: (X, Y)
    :param start_date: Start date as a datetime object. (Earlier date)
    :param end_date: End date as a datetime object. (Later date)
    :param cloud_min: Minimum cloud cover (Default: 0)
    :param cloud_max: Maximum cloud cover.
    :param sort_date: Sort the response by the acquisition date
    :param sort_desc: Sort order (ascending or descending). Ascending: oldest version.
                      Descending: newest version.
    :param page_size: The number of records to be returned by a single query as a
                      'page'.
    :param page_num: The page number to be retrieved allowing results greater than
                     the number which will fit on a single page to be retrieved.
    :param other_params: A dict of other parameters where the key is the search
                         parameter name and the value is the value to search with.
    :return: A list of dictionaries with a dictionary for item.
    """

    prod_srch_params = {"short_name": prod_short_name, "version": version}
    header_info = {"Accept": "application/json"}

    prod_srch_params["page_size"] = page_size
    prod_srch_params["page_num"] = page_num

    if only_dnwld:
        prod_srch_params["downloadable"] = True

    if (start_date is not None) or (end_date is not None):
        start_date_str = ""
        if start_date is not None:
            start_date_str = start_date.strftime("%Y-%m-%dT00:00:00Z")

        end_date_str = ""
        if end_date is not None:
            end_date_str = end_date.strftime("%Y-%m-%dT00:00:00Z")

        prod_srch_params["temporal"] = f"{start_date_str},{end_date_str}"

    if (cloud_min > 0) or (cloud_max is not None):
        cloud_min_str = ""
        if cloud_min > 0:
            cloud_min_str = f"{cloud_min}"

        cloud_max_str = ""
        if cloud_max is not None:
            cloud_max_str = f"{cloud_max}"

        prod_srch_params["cloud_cover"] = f"{cloud_min_str},{cloud_max_str}"

    if pt is not None:
        pt_str = "{},{}".format(pt[0], pt[1])
        prod_srch_params["point"] = pt_str
    elif bbox is not None:
        bbox_str = "{},{},{},{}".format(bbox[0], bbox[2], bbox[1], bbox[3])
        prod_srch_params["bounding_box"] = bbox_str

    if sort_date:
        if sort_desc:
            prod_srch_params["sort_key"] = "-start_date"
        else:
            prod_srch_params["sort_key"] = "%2Bstart_date"

    if other_params is not None:
        prod_srch_params = {**prod_srch_params, **other_params}

    rtnd_data = rsgislib.tools.httptools.send_http_json_request(
        CMR_GRANULES_URL,
        data=prod_srch_params,
        convert_to_json=False,
        header_data=header_info,
    )
    granules_lst = _check_cmr_response(rtnd_data)

    return granules_lst


def find_all_granules(
    prod_short_name: str,
    version: str,
    only_dnwld: bool = True,
    bbox: List[float] = None,
    pt: List[float] = None,
    start_date: datetime.datetime = None,
    end_date: datetime.datetime = None,
    cloud_min: int = 0,
    cloud_max: int = None,
    sort_date: bool = True,
    sort_desc: bool = True,
    page_size: int = 100,
    max_n_pages: int = 100,
    other_params: Dict[str, str] = None,
) -> List[Dict]:
    """
    A function which will find granules from the CMR system for the product of
    interest using the search parameters provided using the find_granules function
    but iterates through all the pages available to return all the available
    granules rather than just a single page.

    :param prod_short_name: the product short name for the product of interest.
    :param version: the version of the product to be retrieved.
    :param only_dnwld: If true (default)
    :param bbox: (MinX, MaxX, MinY, MaxY)
    :param pt: (X, Y)
    :param start_date: Start date as a datetime object. (Earlier date)
    :param end_date: End date as a datetime object. (Later date)
    :param cloud_min: Minimum cloud cover (Default: 0)
    :param cloud_max: Maximum cloud cover.
    :param sort_date: Sort the response by the acquisition date
    :param sort_desc: Sort order (ascending or descending). Ascending: oldest version.
                      Descending: newest version.
    :param page_size: The number of records to be returned by a single query as a
                      'page'. (Default: 100)
    :param max_n_pages: the maximum number of pages returned (Default: 100)
    :param other_params: A dict of other parameters where the key is the search
                         parameter name and the value is the value to search with.
    :return: A list of dictionaries with a dictionary for item.
    """
    import tqdm

    rtn_granules = list()
    print("Note. Progress bar is for the maximum number of pages...")
    print("Hopefully you do not need the maximum so it will be quicker...")
    for i in tqdm.tqdm(range(max_n_pages)):
        page_granules = find_granules(
            prod_short_name=prod_short_name,
            version=version,
            only_dnwld=only_dnwld,
            bbox=bbox,
            pt=pt,
            start_date=start_date,
            end_date=end_date,
            cloud_min=cloud_min,
            cloud_max=cloud_max,
            sort_date=sort_date,
            sort_desc=sort_desc,
            page_size=page_size,
            page_num=i + 1,
            other_params=other_params,
        )
        if len(page_granules) > 0:
            rtn_granules += page_granules
        else:
            break

    return rtn_granules


def get_total_file_size(granule_lst: List[Dict]) -> float:
    """
    A function which using the list granules to sum the total file
    size of the granules in the list. The file size units are whatever
    has been use for the product but seems to be usually be MegaBytes (MB).

    :param granule_lst: List of granules from find_granules or find_all_granules
    :return: float for the total file size.

    """
    tot_file_size = 0.0
    for granule in granule_lst:
        tot_file_size += float(granule["granule_size"])
    return tot_file_size


def create_cmr_dwnld_db(
    db_json: str, granule_lst: List[Dict], dwnld_file_mime_type: str
) -> List[str]:
    """
    A function which iterates through a granule list and builds a json database
    of file to be downloaded. The database can then be used to keep track of
    which files have been successfully downloaded allowing those which haven't
    downloaded to be tried again.

    For the mine types, print a few of the granule list of the terminal (i.e.,
    pprint.pprint(granule_lst) and check what the mime type is for the download
    you want. However, for GEDI data it is probably application/x-hdfeos. 

    :param db_json: The file path for the databases JSON file.
    :param granule_lst: List of granules from find_granules or find_all_granules
    :param dwnld_file_mime_type: (e.g., application/x-hdfeos, application/x-hdf)
    :return: a List of producer_granule_id's for the granules where a
             URL could not be found.

    """
    import pysondb
    import tqdm

    db_data = []
    granules_no_url = []
    for granule in tqdm.tqdm(granule_lst):
        id_val = granule["id"]
        producer_granule_id = granule["producer_granule_id"]
        granule_size = granule["granule_size"]
        found_url = False
        for link in granule["links"]:
            if "type" in link:
                if link["type"] == dwnld_file_mime_type:
                    granule_url = link["href"]
                    found_url = True
                    break
        if found_url:
            db_data.append(
                {
                    "id": id_val,
                    "producer_granule_id": producer_granule_id,
                    "granule_size": granule_size,
                    "granule_url": granule_url,
                    "lcl_path": "",
                    "downloaded": False,
                }
            )
        else:
            granules_no_url.append(producer_granule_id)

    lst_db = pysondb.getDb(db_json)
    if len(db_data) > 0:
        lst_db.addMany(db_data)

    return granules_no_url


def cmr_download_file_http(
    input_url: str,
    out_file_path: str,
    username: str,
    password: str,
    no_except: bool = True,
) -> bool:
    """

    :param input_url: The input remote URL to be downloaded.
    :param out_file_path: the local file path and file name
    :param username: the username for the server
    :param password: the password for the server
    :return: boolean as to whether the file was successfully downloaded or not.

    """
    import tqdm
    import rsgislib.tools.httptools

    session_http = SessionWithHeaderNASARedirection(username, password)

    user_agent = "rsgislib/{}".format(rsgislib.get_rsgislib_version())
    session_http.headers["User-Agent"] = user_agent

    tmp_dwnld_path = out_file_path + ".incomplete"

    headers = {}

    try:
        with session_http.get(input_url, stream=True, headers=headers) as r:
            if rsgislib.tools.httptools.check_http_response(r, input_url):
                total = int(r.headers.get("content-length", 0))
                chunk_size = 2 ** 20
                n_chunks = int(total / chunk_size) + 1

                with open(tmp_dwnld_path, "wb") as f:
                    for chunk in tqdm.tqdm(
                        r.iter_content(chunk_size=chunk_size), total=n_chunks
                    ):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)
        if os.path.exists(tmp_dwnld_path):
            os.rename(tmp_dwnld_path, out_file_path)
            print("Download Complete: {}".format(out_file_path))

    except Exception as e:
        if no_except:
            print(e)
        else:
            raise rsgislib.RSGISPyException("{}".format(e))
        return False
    return True


def download_granules_use_dwnld_db(
    db_json: str, out_path: str, user_pass_file: str, use_wget: bool = False
):
    """
    A function which can use the JSON database built by create_cmr_dwnld_db to
    batch download a set of files keeping track of those which where successfully
    downloaded and those that were unsuccessful.

    :param db_json: file path for the JSON db file.
    :param out_path: the output path where data should be downloaded to.
    :param user_pass_file: path to an encoded (base64) username/password file
    :param use_wget: boolean as to whether to use wget to download files or
                     a pure python function. (Default: False - i.e., pure python).

    """
    import pysondb
    import rsgislib.tools.utils
    import rsgislib.tools.httptools

    username, password = rsgislib.tools.utils.get_username_password(user_pass_file)

    lst_db = pysondb.getDb(db_json)

    dwld_files = lst_db.getByQuery({"downloaded": False})

    for dwn_file in dwld_files:
        out_file_path = os.path.join(out_path, dwn_file["producer_granule_id"])
        print(f"Downloading: {out_file_path}")
        dwn_success = False
        if use_wget:
            dwn_success, dwn_message = rsgislib.tools.httptools.wget_download_file(
                input_url=dwn_file["granule_url"],
                out_file_path=out_file_path,
                username=username,
                password=password,
            )
        else:
            dwn_success = cmr_download_file_http(
                input_url=dwn_file["granule_url"],
                out_file_path=out_file_path,
                username=username,
                password=password,
            )

        if dwn_success:
            lst_db.updateById(
                dwn_file["id"], {"lcl_path": out_file_path, "downloaded": True}
            )



#!/usr/bin/env python
"""
This namespace contains rsgislib Python bindings 

Please be aware that the following variables have been 
defined to match enums within RSGISLib.

Data Types for images:

    * TYPE_UNDEFINED = 0
    * TYPE_8INT = 1
    * TYPE_16INT = 2
    * TYPE_32INT = 3
    * TYPE_64INT = 4
    * TYPE_8UINT = 5
    * TYPE_16UINT = 6
    * TYPE_32UINT = 7
    * TYPE_64UINT = 8
    * TYPE_32FLOAT = 9
    * TYPE_64FLOAT = 10

Methods for the Maximum Likelihood Classifier:

    * METHOD_SAMPLES = 0        # as calculated by ML
    * METHOD_AREA = 1           # priors set by the relative area
    * METHOD_EQUAL = 2          # priors all equal
    * METHOD_USERDEFINED = 3    # priors passed in to function
    * METHOD_WEIGHTED = 4       # priors by area but with a weight applied

Shape indexes used with RasterGIS:

    * SHAPE_SHAPENA = 0
    * SHAPE_SHAPEAREA = 1
    * SHAPE_ASYMMETRY = 2
    * SHAPE_BORDERINDEX = 3
    * SHAPE_BORDERLENGTH = 4
    * SHAPE_COMPACTNESS = 5
    * SHAPE_DENSITY = 6
    * SHAPE_ELLIPTICFIT = 7
    * SHAPE_LENGTH = 8
    * SHAPE_LENGTHWIDTH = 9
    * SHAPE_WIDTH = 10
    * SHAPE_MAINDIRECTION = 11
    * SHAPE_RADIUSLARGESTENCLOSEDELLIPSE = 12
    * SHAPE_RADIUSSMALLESTENCLOSEDELLIPSE = 13
    * SHAPE_RECTANGULARFIT = 14
    * SHAPE_ROUNDNESS = 15
    * SHAPE_SHAPEINDEX = 16

Methods of initialising KMEANS:

    * INITCLUSTER_RANDOM = 0
    * INITCLUSTER_DIAGONAL_FULL = 1
    * INITCLUSTER_DIAGONAL_STDDEV = 2
    * INITCLUSTER_DIAGONAL_FULL_ATTACH = 3
    * INITCLUSTER_DIAGONAL_STDDEV_ATTACH = 4
    * INITCLUSTER_KPP = 5


Methods of calculating distance:

    * DIST_UNDEFINED = 0
    * DIST_EUCLIDEAN = 1
    * DIST_MAHALANOBIS = 2
    * DIST_MANHATTEN = 3
    * DIST_MINKOWSKI = 4
    * DIST_CHEBYSHEV = 5
    * DIST_MUTUALINFO = 6

Methods of summarising data:

    * SUMTYPE_UNDEFINED = 0
    * SUMTYPE_MODE = 1
    * SUMTYPE_MEAN = 2
    * SUMTYPE_MEDIAN = 3
    * SUMTYPE_MIN = 4
    * SUMTYPE_MAX = 5
    * SUMTYPE_STDDEV = 6
    * SUMTYPE_COUNT = 7
    * SUMTYPE_RANGE = 8
    * SUMTYPE_SUM = 9

Constants specifying how bands should be treated when sharpening
(see rsgislib.imageutils)

    * SHARP_RES_IGNORE = 0
    * SHARP_RES_LOW = 1
    * SHARP_RES_HIGH = 2

Options for interpolating raster data:

    * INTERP_NEAREST_NEIGHBOUR = 0
    * INTERP_BILINEAR = 1
    * INTERP_CUBIC = 2
    * INTERP_CUBICSPLINE = 3
    * INTERP_LANCZOS = 4
    * INTERP_AVERAGE = 5
    * INTERP_MODE = 6

"""
from __future__ import print_function

import time
import datetime
import sys
import pathlib

from osgeo import gdal

gdal.UseExceptions()

TYPE_UNDEFINED = 0
TYPE_8INT = 1
TYPE_16INT = 2
TYPE_32INT = 3
TYPE_64INT = 4
TYPE_8UINT = 5
TYPE_16UINT = 6
TYPE_32UINT = 7
TYPE_64UINT = 8
TYPE_32FLOAT = 9
TYPE_64FLOAT = 10

DIST_UNDEFINED = 0
DIST_EUCLIDEAN = 1
DIST_MAHALANOBIS = 2
DIST_MANHATTEN = 3
DIST_MINKOWSKI = 4
DIST_CHEBYSHEV = 5
DIST_MUTUALINFO = 6

SUMTYPE_UNDEFINED = 0
SUMTYPE_MODE = 1
SUMTYPE_MEAN = 2
SUMTYPE_MEDIAN = 3
SUMTYPE_MIN = 4
SUMTYPE_MAX = 5
SUMTYPE_STDDEV = 6
SUMTYPE_COUNT = 7
SUMTYPE_RANGE = 8
SUMTYPE_SUM = 9

METHOD_SAMPLES = 0  # as calculated by ML
METHOD_AREA = 1  # priors set by the relative area
METHOD_EQUAL = 2  # priors all equal
METHOD_USERDEFINED = 3  # priors passed in to function
METHOD_WEIGHTED = 4  # priors by area but with a weight applied

SHAPE_SHAPENA = 0
SHAPE_SHAPEAREA = 1
SHAPE_ASYMMETRY = 2
SHAPE_BORDERINDEX = 3
SHAPE_BORDERLENGTH = 4
SHAPE_COMPACTNESS = 5
SHAPE_DENSITY = 6
SHAPE_ELLIPTICFIT = 7
SHAPE_LENGTH = 8
SHAPE_LENGTHWIDTH = 9
SHAPE_WIDTH = 10
SHAPE_MAINDIRECTION = 11
SHAPE_RADIUSLARGESTENCLOSEDELLIPSE = 12
SHAPE_RADIUSSMALLESTENCLOSEDELLIPSE = 13
SHAPE_RECTANGULARFIT = 14
SHAPE_ROUNDNESS = 15
SHAPE_SHAPEINDEX = 16

INITCLUSTER_RANDOM = 0
INITCLUSTER_DIAGONAL_FULL = 1
INITCLUSTER_DIAGONAL_STDDEV = 2
INITCLUSTER_DIAGONAL_FULL_ATTACH = 3
INITCLUSTER_DIAGONAL_STDDEV_ATTACH = 4
INITCLUSTER_KPP = 5

SHARP_RES_IGNORE = 0
SHARP_RES_LOW = 1
SHARP_RES_HIGH = 2

INTERP_NEAREST_NEIGHBOUR = 0
INTERP_BILINEAR = 1
INTERP_CUBIC = 2
INTERP_CUBICSPLINE = 3
INTERP_LANCZOS = 4
INTERP_AVERAGE = 5
INTERP_MODE = 6

GEOM_PT = 1
GEOM_LINE = 2
GEOM_POLY = 3
GEOM_MLINE = 4
GEOM_MPOLY = 5


def get_install_base_path() -> pathlib.PurePath:
    """
    A function which returns the base path for the RSGISLib installation.
    This can be used to find files stored in share etc. installed alongside
    the RSGISLib source code.

    :return: a pathlib.PurePath object with the base path of the RSGISLib
             installation.

    """
    import site
    import platform

    site_path_str = site.getsitepackages()[0]
    site_path = pathlib.PurePath(site_path_str)

    base_path_comps = []
    for path_part in site_path.parts:
        if path_part != "lib":
            base_path_comps.append(path_part)
        else:
            break

    if platform.system() == "Windows":
        base_path = pathlib.PureWindowsPath(*base_path_comps)
    else:
        base_path = pathlib.PurePosixPath(*base_path_comps)

    return base_path


RSGISLIB_INSTALL_PREFIX = str(get_install_base_path())


def get_rsgislib_version() -> str:
    """
    Calls rsgis-config to get the version number.

    :return: string with the rsgislib version number

    """

    # Try calling rsgis-config to get minor version number
    try:
        import distutils.spawn

        if distutils.spawn.find_executable("rsgis-config") is not None:
            import subprocess

            out = subprocess.run(
                ["rsgis-config", "--version"],
                capture_output=True,
                text=True,
                check=True,
            )
            version_str = out.stdout
            version_str = version_str.split("\n")[0]
        else:
            version_str = "NA.NA"
    except Exception:
        version_str = "NA.NA"
    return version_str


__version__ = get_rsgislib_version()

py_sys_version_str = "{}.{}".format(sys.version_info.major, sys.version_info.minor)
py_sys_version_flt = float(py_sys_version_str)


class RSGISPyException(Exception):
    """
    A class representing the RSGISLib exception.
    """

    def __init__(self, value: str):
        """
        Init for the RSGISPyException class
        """
        self.value = value

    def __str__(self):
        """
        Return a string representation of the exception
        """
        return repr(self.value)


class RSGISGDALErrorHandler:
    """
    A class representing a generic GDAL Error Handler which
    can be used to pick up GDAL warnings rather than just
    failure errors.
    """

    def __init__(self):
        """
        Init for RSGISGDALErrorHandler. Class attributes are
        err_level, err_no and err_msg

        """
        self.err_level = gdal.CE_None
        self.err_no = 0
        self.err_msg = ""

    def handler(self, err_level, err_no, err_msg):
        """
        The handler function which is called with the error information.

        :param err_level: The level of the error
        :param err_no: The error number
        :param err_msg: The message (string) associated with the error.

        """
        self.err_level = err_level
        self.err_no = err_no
        self.err_msg = err_msg


def get_rsgislib_datatype(gdal_type: int) -> int:
    """
    Convert from GDAL data type string to RSGISLib data type int.

    :return: int

    """
    gdal_type = gdal_type.lower()
    if gdal_type == "int8":
        out_dt = TYPE_8INT
    elif gdal_type == "int16":
        out_dt = TYPE_16INT
    elif gdal_type == "int32":
        out_dt = TYPE_32INT
    elif gdal_type == "int64":
        out_dt = TYPE_64INT
    elif gdal_type == "byte" or gdal_type == "uint8":
        out_dt = TYPE_8UINT
    elif gdal_type == "uint16":
        out_dt = TYPE_16UINT
    elif gdal_type == "uint32":
        out_dt = TYPE_32UINT
    elif gdal_type == "uint64":
        out_dt = TYPE_64UINT
    elif gdal_type == "float32":
        out_dt = TYPE_32FLOAT
    elif gdal_type == "float64":
        out_dt = TYPE_64FLOAT
    else:
        raise RSGISPyException(
            f"The data type '{gdal_type}' is unknown / not supported."
        )
    return out_dt


def get_gdal_datatype(rsgislib_datatype: int) -> int:
    """
    Convert from RSGIS data type to GDAL data type int.

    :return: int

    """
    if rsgislib_datatype == TYPE_16INT:
        out_dt = gdal.GDT_Int16
    elif rsgislib_datatype == TYPE_32INT:
        out_dt = gdal.GDT_Int32
    elif rsgislib_datatype == TYPE_8UINT:
        out_dt = gdal.GDT_Byte
    elif rsgislib_datatype == TYPE_16UINT:
        out_dt = gdal.GDT_UInt16
    elif rsgislib_datatype == TYPE_32UINT:
        out_dt = gdal.GDT_UInt32
    elif rsgislib_datatype == TYPE_32FLOAT:
        out_dt = gdal.GDT_Float32
    elif rsgislib_datatype == TYPE_64FLOAT:
        out_dt = gdal.GDT_Float64
    else:
        raise RSGISPyException(
            f"The data type '{rsgislib_datatype}' is unknown / not supported."
        )
    return out_dt


def get_numpy_datatype(rsgislib_datatype: int) -> int:
    """
    Convert from RSGISLib data type to numpy datatype

    :param rsgis_datatype:
    :return: numpy datatype

    """
    import numpy

    if rsgislib_datatype == TYPE_8INT:
        numpy_dt = numpy.int8
    elif rsgislib_datatype == TYPE_16INT:
        numpy_dt = numpy.int16
    elif rsgislib_datatype == TYPE_32INT:
        numpy_dt = numpy.int32
    elif rsgislib_datatype == TYPE_64INT:
        numpy_dt = numpy.int64
    elif rsgislib_datatype == TYPE_8UINT:
        numpy_dt = numpy.uint8
    elif rsgislib_datatype == TYPE_16UINT:
        numpy_dt = numpy.uint16
    elif rsgislib_datatype == TYPE_32UINT:
        numpy_dt = numpy.uint32
    elif rsgislib_datatype == TYPE_64UINT:
        numpy_dt = numpy.uint64
    elif rsgislib_datatype == TYPE_32FLOAT:
        numpy_dt = numpy.float32
    elif rsgislib_datatype == TYPE_64FLOAT:
        numpy_dt = numpy.float64
    else:
        raise RSGISPyException("Datatype was not recognised.")
    return numpy_dt


def get_numpy_char_codes_datatype(rsgislib_datatype: int) -> str:
    """
    Convert from RSGISLib data type to numpy datatype

    :param rsgis_datatype:
    :return: numpy character code datatype

    """
    import numpy

    if rsgislib_datatype == TYPE_8INT:
        numpy_dt = numpy.dtype(numpy.int8).char
    elif rsgislib_datatype == TYPE_16INT:
        numpy_dt = numpy.dtype(numpy.int16).char
    elif rsgislib_datatype == TYPE_32INT:
        numpy_dt = numpy.dtype(numpy.int32).char
    elif rsgislib_datatype == TYPE_64INT:
        numpy_dt = numpy.dtype(numpy.int64).char
    elif rsgislib_datatype == TYPE_8UINT:
        numpy_dt = numpy.dtype(numpy.uint8).char
    elif rsgislib_datatype == TYPE_16UINT:
        numpy_dt = numpy.dtype(numpy.uint16).char
    elif rsgislib_datatype == TYPE_32UINT:
        numpy_dt = numpy.dtype(numpy.uint32).char
    elif rsgislib_datatype == TYPE_64UINT:
        numpy_dt = numpy.dtype(numpy.uint64).char
    elif rsgislib_datatype == TYPE_32FLOAT:
        numpy_dt = numpy.dtype(numpy.float32).char
    elif rsgislib_datatype == TYPE_64FLOAT:
        numpy_dt = numpy.dtype(numpy.float64).char
    else:
        raise RSGISPyException("Datatype was not recognised.")
    return numpy_dt


def is_notebook():
    """
    A function to find if running within a python notebook. If
    running within a jupyter notebook then can use a different
    progress bar.

    :return: boolean: True: within notebook

    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


class RSGISTime:
    """
    Class to calculate run time for a function, format and print out.

    Need to call start before running function and end immediately after.

    .. code:: python

        t = RSGISTime()
        t.start()
        rsgislib.segmentation.clump(kMeansFileZonesNoSgls, initClumpsFile,
                                    gdalformat, False, 0)
        t.end()

    Note, this is only designed to provide some general feedback, for
    benchmarking the timeit module is better suited.

    """

    def __init__(self):
        self.start_time = time.time()
        self.end_time = time.time()

    def start(self, print_start_time: bool = False):
        """
        Start timer, optionally printing start time

        :param print_start_time: A boolean specifying whether the start time
                                 should be printed to console.

        """
        self.start_time = time.time()
        if print_start_time:
            print(time.strftime("Start Time: %H:%M:%S, %a %b %m %Y."))

    def end(self, report_diff: bool = True, precede_str: str = "", post_str: str = ""):
        """
        End timer and optionally print difference.
        If precedeStr or postStr have a value then they will be used instead
        of the generic wording around the time.

        "{precede_str} {time} {post_str}"

        :param report_diff: A boolean specifying whether time difference should
                            be printed to console.
        :param precede_str: A string which is printed ahead of time difference
        :param post_str: A string which is printed after the time difference

        """
        self.end_time = time.time()
        if report_diff:
            self.calc_diff(precede_str, post_str)

    def calc_diff(self, precede_str: str = "", post_str: str = ""):
        """
        Calculate time difference, format and print.

        :param precede_str: A string which is printed ahead of time difference
        :param post_str: A string which is printed after the time difference

        """
        time_diff = self.end_time - self.start_time

        use_custom_mss = False
        if (len(precede_str) > 0) or (len(post_str) > 0):
            use_custom_mss = True

        if time_diff <= 1:
            if use_custom_mss:
                print(f"{precede_str} in less than a second {post_str}")
            else:
                print("Algorithm Completed in less than a second.")
        else:
            time_obj = datetime.datetime.utcfromtimestamp(time_diff)
            time_diff_str = time_obj.strftime("%H:%M:%S")
            if use_custom_mss:
                print(f"{precede_str} {time_diff_str} {post_str}")
            else:
                print(f"Algorithm Completed in {time_diff_str}.")


class TQDMProgressBar:
    """
    Uses TQDM TermProgress to print a progress bar to the terminal
    """

    def __init__(self):
        self.lprogress = 0

    def setTotalSteps(self, steps: int):
        import tqdm

        if is_notebook():
            import tqdm.notebook

            self.pbar = tqdm.notebook.tqdm(total=steps)
        else:
            self.pbar = tqdm.tqdm(total=steps)
        self.lprogress = 0

    def setProgress(self, progress: int):
        step = progress - self.lprogress
        self.pbar.update(step)
        self.lprogress = progress

    def reset(self):
        self.pbar.close()
        import tqdm

        if is_notebook():
            import tqdm.notebook

            self.pbar = tqdm.notebook.tqdm(total=100)
        else:
            self.pbar = tqdm.tqdm(total=100)
        self.lprogress = 0

    def setLabelText(self, text: str):
        sys.stdout.write("\n%s\n" % text)

    def wasCancelled(self):
        return False

    def displayException(self, trace):
        sys.stdout.write(trace)

    def displayWarning(self, text: str):
        sys.stdout.write("Warning: %s\n" % text)

    def displayError(self, text: str):
        sys.stdout.write("Error: %s\n" % text)

    def displayInfo(self, text: str):
        sys.stdout.write("Info: %s\n" % text)



#! /usr/bin/env python
############################################################################
#  classxgboost.py
#
#  Copyright 2020 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 16/02/2020
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

from __future__ import print_function

import rsgislib
import rsgislib.imageutils
import rsgislib.imagecalc
import rsgislib.rastergis

import numpy

import h5py

from osgeo import gdal

from rios import applier
from rios import cuiprogress
from rios import rat

HAVE_XGBOOST = True
try:
    import xgboost as xgb
except ImportError:
    HAVE_XGBOOST = False

from sklearn.metrics import roc_auc_score
from sklearn.metrics import accuracy_score

import gc

import json


def optimise_xgboost_binary_classifier(
    out_params_file,
    cls1_train_file,
    cls1_valid_file,
    cls2_train_file,
    cls2_valid_file,
    n_threads=1,
    scale_pos_weight=None,
    mdl_cls_obj=None,
):
    """
    A function which performs a bayesian optimisation of the hyper-parameters for a binary xgboost
    classifier. Class 1 is the class which you are interested in and Class 2 is the 'other class'.

    This function requires that xgboost and skopt modules to be installed.

    :param out_params_file: The output model parameters which have been optimised.
    :param cls1_train_file: Training samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_valid_file: Validation samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_test_file: Testing samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls2_train_file: Training samples HDF5 file for the 'other' class
    :param cls2_valid_file: Validation samples HDF5 file for the 'other' class
    :param cls2_test_file: Testing samples HDF5 file for the 'other' class
    :param n_threads: The number of threads to use for the training.
    :param scale_pos_weight: Optional, default is None. If None then a value will automatically be calculated.
                             Parameter used to balance imbalanced training data.
    :param mdl_cls_obj: XGBoost object to allow continue training with a new dataset.

    """
    if not HAVE_XGBOOST:
        raise rsgislib.RSGISPyException("Do not have xgboost module installed.")

    from skopt.space import Real, Integer
    from skopt import gp_minimize

    print("Reading Class 1 Training")
    f = h5py.File(cls1_train_file, "r")
    num_cls1_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_train_rows = {}".format(num_cls1_train_rows))
    train_cls1 = numpy.array(f["DATA/DATA"])
    train_cls1_lbl = numpy.ones(num_cls1_train_rows, dtype=int)

    print("Reading Class 1 Validation")
    f = h5py.File(cls1_valid_file, "r")
    num_cls1_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_valid_rows = {}".format(num_cls1_valid_rows))
    valid_cls1 = numpy.array(f["DATA/DATA"])
    valid_cls1_lbl = numpy.ones(num_cls1_valid_rows, dtype=int)

    print("Reading Class 2 Training")
    f = h5py.File(cls2_train_file, "r")
    num_cls2_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_train_rows = {}".format(num_cls2_train_rows))
    train_cls2 = numpy.array(f["DATA/DATA"])
    train_cls2_lbl = numpy.zeros(num_cls2_train_rows, dtype=int)

    print("Reading Class 2 Validation")
    f = h5py.File(cls2_valid_file, "r")
    num_cls2_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_valid_rows = {}".format(num_cls2_valid_rows))
    valid_cls2 = numpy.array(f["DATA/DATA"])
    valid_cls2_lbl = numpy.zeros(num_cls2_valid_rows, dtype=int)

    print("Finished Reading Data")

    vaild_np = numpy.concatenate((valid_cls2, valid_cls1))
    vaild_lbl_np = numpy.concatenate((valid_cls2_lbl, valid_cls1_lbl))
    d_valid = xgb.DMatrix(vaild_np, label=vaild_lbl_np)

    d_train = xgb.DMatrix(
        numpy.concatenate((train_cls2, train_cls1)),
        label=numpy.concatenate((train_cls2_lbl, train_cls1_lbl)),
    )

    space = [
        Real(0.01, 0.9, name="eta"),
        Integer(0, 100, name="gamma"),
        Integer(2, 20, name="max_depth"),
        Integer(1, 10, name="min_child_weight"),
        Integer(0, 10, name="max_delta_step"),
        Real(0.5, 1, name="subsample"),
        Integer(2, 100, name="num_boost_round"),
    ]

    if scale_pos_weight is None:
        scale_pos_weight = num_cls2_train_rows / num_cls1_train_rows
        if scale_pos_weight < 1:
            scale_pos_weight = 1
    print("scale_pos_weight = {}".format(scale_pos_weight))

    def _objective(values):
        params = {
            "eta": values[0],
            "gamma": values[1],
            "max_depth": values[2],
            "min_child_weight": values[3],
            "max_delta_step": values[4],
            "subsample": values[5],
            "nthread": n_threads,
            "eval_metric": "auc",
            "objective": "binary:logistic",
        }

        print("\nNext set of params.....", params)

        num_boost_round = values[6]
        print("num_boost_round = {}.".format(num_boost_round))

        watchlist = [(d_train, "train"), (d_valid, "validation")]
        evals_results = {}
        model_xgb = xgb.train(
            params,
            d_train,
            num_boost_round,
            evals=watchlist,
            evals_result=evals_results,
            verbose_eval=False,
            xgb_model=mdl_cls_obj,
        )

        auc = -roc_auc_score(vaild_lbl_np, model_xgb.predict(d_valid))
        print("\nAUROC.....", -auc, ".....iter.....")
        gc.collect()
        return auc

    res_gp = gp_minimize(
        _objective, space, n_calls=20, random_state=0, n_random_starts=10
    )

    print("Best score={}".format(res_gp.fun))

    best_params = res_gp.x

    params = {
        "eta": float(best_params[0]),
        "gamma": int(best_params[1]),
        "max_depth": int(best_params[2]),
        "min_child_weight": int(best_params[3]),
        "max_delta_step": int(best_params[4]),
        "subsample": float(best_params[5]),
        "nthread": int(n_threads),
        "eval_metric": "auc",
        "objective": "binary:logistic",
        "num_boost_round": int(best_params[6]),
    }

    with open(out_params_file, "w") as fp:
        json.dump(
            params,
            fp,
            sort_keys=True,
            indent=4,
            separators=(",", ": "),
            ensure_ascii=False,
        )


def train_xgboost_binary_classifier(
    out_mdl_file,
    cls_params_file,
    cls1_train_file,
    cls1_valid_file,
    cls1_test_file,
    cls2_train_file,
    cls2_valid_file,
    cls2_test_file,
    n_threads=1,
    mdl_cls_obj=None,
):
    """
    A function which performs a bayesian optimisation of the hyper-parameters for a binary xgboost
    classifier. Class 1 is the class which you are interested in and Class 2 is the 'other class'.

    This function requires that xgboost and skopt modules to be installed.

    :param out_mdl_file: The output model which can be loaded to perform a classification.
    :param cls_params_file: A JSON file with the model parameters
    :param cls1_train_file: Training samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_valid_file: Validation samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_test_file: Testing samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls2_train_file: Training samples HDF5 file for the 'other' class
    :param cls2_valid_file: Validation samples HDF5 file for the 'other' class
    :param cls2_test_file: Testing samples HDF5 file for the 'other' class
    :param n_threads: The number of threads to use for the training.
    :param scale_pos_weight: Optional, default is None. If None then a value will automatically be calculated.
                             Parameter used to balance imbalanced training data.
    :param mdl_cls_obj: XGBoost object to allow continue training with a new dataset.

    """
    if not HAVE_XGBOOST:
        raise rsgislib.RSGISPyException("Do not have xgboost module installed.")

    print("Reading Class 1 Training")
    f = h5py.File(cls1_train_file, "r")
    num_cls1_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_train_rows = {}".format(num_cls1_train_rows))
    train_cls1 = numpy.array(f["DATA/DATA"])
    train_cls1_lbl = numpy.ones(num_cls1_train_rows, dtype=int)

    print("Reading Class 1 Validation")
    f = h5py.File(cls1_valid_file, "r")
    num_cls1_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_valid_rows = {}".format(num_cls1_valid_rows))
    valid_cls1 = numpy.array(f["DATA/DATA"])
    valid_cls1_lbl = numpy.ones(num_cls1_valid_rows, dtype=int)

    print("Reading Class 1 Testing")
    f = h5py.File(cls1_test_file, "r")
    num_cls1_test_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_test_rows = {}".format(num_cls1_test_rows))
    test_cls1 = numpy.array(f["DATA/DATA"])
    test_cls1_lbl = numpy.ones(num_cls1_test_rows, dtype=int)

    print("Reading Class 2 Training")
    f = h5py.File(cls2_train_file, "r")
    num_cls2_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_train_rows = {}".format(num_cls2_train_rows))
    train_cls2 = numpy.array(f["DATA/DATA"])
    train_cls2_lbl = numpy.zeros(num_cls2_train_rows, dtype=int)

    print("Reading Class 2 Validation")
    f = h5py.File(cls2_valid_file, "r")
    num_cls2_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_valid_rows = {}".format(num_cls2_valid_rows))
    valid_cls2 = numpy.array(f["DATA/DATA"])
    valid_cls2_lbl = numpy.zeros(num_cls2_valid_rows, dtype=int)

    print("Reading Class 2 Testing")
    f = h5py.File(cls2_test_file, "r")
    num_cls2_test_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_test_rows = {}".format(num_cls2_test_rows))
    test_cls2 = numpy.array(f["DATA/DATA"])
    test_cls2_lbl = numpy.zeros(num_cls2_test_rows, dtype=int)

    print("Finished Reading Data")

    vaild_np = numpy.concatenate((valid_cls2, valid_cls1))
    vaild_lbl_np = numpy.concatenate((valid_cls2_lbl, valid_cls1_lbl))
    d_valid = xgb.DMatrix(vaild_np, label=vaild_lbl_np)

    d_train = xgb.DMatrix(
        numpy.concatenate((train_cls2, train_cls1)),
        label=numpy.concatenate((train_cls2_lbl, train_cls1_lbl)),
    )

    test_np = numpy.concatenate((test_cls2, test_cls1))
    test_lbl_np = numpy.concatenate((test_cls2_lbl, test_cls1_lbl))
    d_test = xgb.DMatrix(test_np, label=test_lbl_np)

    with open(cls_params_file) as f:
        cls_params = json.load(f)

    print("Start Training Find Classifier")

    params = {
        "eta": cls_params["eta"],
        "gamma": cls_params["gamma"],
        "max_depth": cls_params["max_depth"],
        "min_child_weight": cls_params["min_child_weight"],
        "max_delta_step": cls_params["max_delta_step"],
        "subsample": cls_params["subsample"],
        "nthread": n_threads,
        "eval_metric": cls_params["eval_metric"],
        "objective": cls_params["objective"],
    }

    num_boost_round = cls_params["num_boost_round"]

    evals_results = {}
    watchlist = [(d_train, "train"), (d_valid, "validation")]
    model = xgb.train(
        params,
        d_train,
        num_boost_round,
        evals=watchlist,
        evals_result=evals_results,
        verbose_eval=False,
        xgb_model=mdl_cls_obj,
    )
    test_auc = roc_auc_score(test_lbl_np, model.predict(d_test))
    print("Testing AUC: {}".format(test_auc))
    print("Finish Training")

    model.save_model(out_mdl_file)

    pred_test = model.predict(d_test)
    for i in range(test_np.shape[0]):
        if pred_test[i] >= 0.5:
            pred_test[i] = 1
        else:
            pred_test[i] = 0
    len(pred_test)

    test_acc = accuracy_score(test_lbl_np, pred_test)
    print("Testing Accuracy: {}".format(test_acc))


def train_opt_xgboost_binary_classifier(
    out_mdl_file,
    cls1_train_file,
    cls1_valid_file,
    cls1_test_file,
    cls2_train_file,
    cls2_valid_file,
    cls2_test_file,
    n_threads=1,
    scale_pos_weight=None,
    mdl_cls_obj=None,
    out_params_file=None,
):
    """
    A function which performs a bayesian optimisation of the hyper-parameters for a binary xgboost
    classifier. Class 1 is the class which you are interested in and Class 2 is the 'other class'.

    This function requires that xgboost and skopt modules to be installed.

    :param out_mdl_file: The output model which can be loaded to perform a classification.
    :param cls1_train_file: Training samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_valid_file: Validation samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_test_file: Testing samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls2_train_file: Training samples HDF5 file for the 'other' class
    :param cls2_valid_file: Validation samples HDF5 file for the 'other' class
    :param cls2_test_file: Testing samples HDF5 file for the 'other' class
    :param n_threads: The number of threads to use for the training.
    :param scale_pos_weight: Optional, default is None. If None then a value will automatically be calculated.
                             Parameter used to balance imbalanced training data.
    :param mdl_cls_obj: XGBoost object to allow continue training with a new dataset.
    :param out_params_file: The output model parameters which have been optimised.
                            If None then no file will be outputted.

    """
    if not HAVE_XGBOOST:
        raise rsgislib.RSGISPyException("Do not have xgboost module installed.")

    from skopt.space import Real, Integer
    from skopt import gp_minimize

    print("Reading Class 1 Training")
    f = h5py.File(cls1_train_file, "r")
    num_cls1_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_train_rows = {}".format(num_cls1_train_rows))
    train_cls1 = numpy.array(f["DATA/DATA"])
    train_cls1_lbl = numpy.ones(num_cls1_train_rows, dtype=int)

    print("Reading Class 1 Validation")
    f = h5py.File(cls1_valid_file, "r")
    num_cls1_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_valid_rows = {}".format(num_cls1_valid_rows))
    valid_cls1 = numpy.array(f["DATA/DATA"])
    valid_cls1_lbl = numpy.ones(num_cls1_valid_rows, dtype=int)

    print("Reading Class 1 Testing")
    f = h5py.File(cls1_test_file, "r")
    num_cls1_test_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_test_rows = {}".format(num_cls1_test_rows))
    test_cls1 = numpy.array(f["DATA/DATA"])
    test_cls1_lbl = numpy.ones(num_cls1_test_rows, dtype=int)

    print("Reading Class 2 Training")
    f = h5py.File(cls2_train_file, "r")
    num_cls2_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_train_rows = {}".format(num_cls2_train_rows))
    train_cls2 = numpy.array(f["DATA/DATA"])
    train_cls2_lbl = numpy.zeros(num_cls2_train_rows, dtype=int)

    print("Reading Class 2 Validation")
    f = h5py.File(cls2_valid_file, "r")
    num_cls2_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_valid_rows = {}".format(num_cls2_valid_rows))
    valid_cls2 = numpy.array(f["DATA/DATA"])
    valid_cls2_lbl = numpy.zeros(num_cls2_valid_rows, dtype=int)

    print("Reading Class 2 Testing")
    f = h5py.File(cls2_test_file, "r")
    num_cls2_test_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_test_rows = {}".format(num_cls2_test_rows))
    test_cls2 = numpy.array(f["DATA/DATA"])
    test_cls2_lbl = numpy.zeros(num_cls2_test_rows, dtype=int)

    print("Finished Reading Data")

    vaild_np = numpy.concatenate((valid_cls2, valid_cls1))
    vaild_lbl_np = numpy.concatenate((valid_cls2_lbl, valid_cls1_lbl))
    d_valid = xgb.DMatrix(vaild_np, label=vaild_lbl_np)

    d_train = xgb.DMatrix(
        numpy.concatenate((train_cls2, train_cls1)),
        label=numpy.concatenate((train_cls2_lbl, train_cls1_lbl)),
    )

    test_np = numpy.concatenate((test_cls2, test_cls1))
    test_lbl_np = numpy.concatenate((test_cls2_lbl, test_cls1_lbl))
    d_test = xgb.DMatrix(test_np, label=test_lbl_np)

    space = [
        Real(0.01, 0.9, name="eta"),
        Integer(0, 100, name="gamma"),
        Integer(2, 20, name="max_depth"),
        Integer(1, 10, name="min_child_weight"),
        Integer(0, 10, name="max_delta_step"),
        Real(0.5, 1, name="subsample"),
        Integer(2, 100, name="num_boost_round"),
    ]

    if scale_pos_weight is None:
        scale_pos_weight = num_cls2_train_rows / num_cls1_train_rows
        if scale_pos_weight < 1:
            scale_pos_weight = 1
    print("scale_pos_weight = {}".format(scale_pos_weight))

    def _objective(values):
        params = {
            "eta": values[0],
            "gamma": values[1],
            "max_depth": values[2],
            "min_child_weight": values[3],
            "max_delta_step": values[4],
            "subsample": values[5],
            "nthread": n_threads,
            "eval_metric": "auc",
            "objective": "binary:logistic",
        }

        print("\nNext set of params.....", params)

        num_boost_round = values[6]
        print("num_boost_round = {}.".format(num_boost_round))

        watchlist = [(d_train, "train"), (d_valid, "validation")]
        evals_results = {}
        model_xgb = xgb.train(
            params,
            d_train,
            num_boost_round,
            evals=watchlist,
            evals_result=evals_results,
            verbose_eval=False,
            xgb_model=mdl_cls_obj,
        )

        auc = -roc_auc_score(vaild_lbl_np, model_xgb.predict(d_valid))
        print("\nAUROC.....", -auc, ".....iter.....")
        gc.collect()
        return auc

    res_gp = gp_minimize(
        _objective, space, n_calls=20, random_state=0, n_random_starts=10
    )

    print("Best score={}".format(res_gp.fun))

    best_params = res_gp.x

    print("Best Params:\n{}".format(best_params))

    if out_params_file is not None:
        out_params = {
            "eta": float(best_params[0]),
            "gamma": int(best_params[1]),
            "max_depth": int(best_params[2]),
            "min_child_weight": int(best_params[3]),
            "max_delta_step": int(best_params[4]),
            "subsample": float(best_params[5]),
            "nthread": int(n_threads),
            "eval_metric": "auc",
            "objective": "binary:logistic",
            "num_boost_round": int(best_params[6]),
        }

        with open(out_params_file, "w") as fp:
            json.dump(
                out_params,
                fp,
                sort_keys=True,
                indent=4,
                separators=(",", ": "),
                ensure_ascii=False,
            )

    print("Start Training Find Classifier")

    params = {
        "eta": best_params[0],
        "gamma": best_params[1],
        "max_depth": best_params[2],
        "min_child_weight": best_params[3],
        "max_delta_step": best_params[4],
        "subsample": best_params[5],
        "nthread": n_threads,
        "eval_metric": "auc",
        "objective": "binary:logistic",
    }

    num_boost_round = best_params[6]

    evals_results = {}
    watchlist = [(d_train, "train"), (d_valid, "validation")]
    model = xgb.train(
        params,
        d_train,
        num_boost_round,
        evals=watchlist,
        evals_result=evals_results,
        verbose_eval=False,
        xgb_model=mdl_cls_obj,
    )
    test_auc = roc_auc_score(test_lbl_np, model.predict(d_test))
    print("Testing AUC: {}".format(test_auc))
    print("Finish Training")

    model.save_model(out_mdl_file)

    pred_test = model.predict(d_test)
    for i in range(test_np.shape[0]):
        if pred_test[i] >= 0.5:
            pred_test[i] = 1
        else:
            pred_test[i] = 0
    len(pred_test)

    test_acc = accuracy_score(test_lbl_np, pred_test)
    print("Testing Accuracy: {}".format(test_acc))


def apply_xgboost_binary_classifier(
    model_file,
    in_msk_img,
    img_mask_val,
    img_file_info,
    out_prob_img,
    gdalformat,
    out_class_img=None,
    class_thres=5000,
    n_threads=1,
):
    """
    This function applies a trained binary (i.e., two classes) xgboost model. The function train_xgboost_binary_classifier
    can be used to train such as model. The output image will contain the probability of membership to the class of
    interest. You will need to threshold this image to get a final hard classification. Alternative, a hard class output
    image and threshold can be applied to this image.

    :param model_file: a trained xgboost binary model which can be loaded with lgb.Booster(model_file=model_file).
    :param in_msk_img: is an image file providing a mask to specify where should be classified. Simplest mask is all the
                    valid data regions (rsgislib.imageutils.gen_valid_mask)
    :param img_mask_val: the pixel value within the imgMask to limit the region to which the classification is applied.
                       Can be used to create a heirachical classification.
    :param img_file_info: a list of rsgislib.imageutils.ImageBandInfo objects (also used within
                        rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) to identify which images and bands are to
                        be used for the classification so it adheres to the training data.
    :param out_prob_img: output image file with the classification probabilities - this image is scaled by
                       multiplying by 10000.
    :param gdalformat: is the output image format - all GDAL supported formats are supported.
    :param out_class_img: Optional output image which will contain the hard classification, defined with a threshold on the
                        probability image.
    :param class_thres: The threshold used to define the hard classification. Default is 5000 (i.e., probability of 0.5).
    :param n_threads: The number of threads to use for the classifier.

    """
    if not HAVE_XGBOOST:
        raise rsgislib.RSGISPyException("Do not have xgboost module installed.")

    def _applyXGBClassifier(info, inputs, outputs, otherargs):
        outClassVals = numpy.zeros_like(inputs.imageMask, dtype=numpy.uint16)
        if numpy.any(inputs.imageMask == otherargs.mskVal):
            outClassVals = outClassVals.flatten()
            imgMaskVals = inputs.imageMask.flatten()
            classVars = numpy.zeros(
                (outClassVals.shape[0], otherargs.numClassVars), dtype=numpy.float32
            )
            # Array index which can be used to populate the output array following masking etc.
            ID = numpy.arange(imgMaskVals.shape[0])
            classVarsIdx = 0
            for imgFile in otherargs.imgFileInfo:
                imgArr = inputs.__dict__[imgFile.name]
                for band in imgFile.bands:
                    classVars[..., classVarsIdx] = imgArr[(band - 1)].flatten()
                    classVarsIdx = classVarsIdx + 1
            classVars = classVars[imgMaskVals == otherargs.mskVal]
            ID = ID[imgMaskVals == otherargs.mskVal]
            predClass = numpy.around(
                otherargs.classifier.predict(xgb.DMatrix(classVars)) * 10000
            )
            outClassVals[ID] = predClass
            outClassVals = numpy.expand_dims(
                outClassVals.reshape(
                    (inputs.imageMask.shape[1], inputs.imageMask.shape[2])
                ),
                axis=0,
            )
        outputs.outimage = outClassVals

    classifier = xgb.Booster({"nthread": n_threads})
    classifier.load_model(model_file)

    infiles = applier.FilenameAssociations()
    infiles.imageMask = in_msk_img
    numClassVars = 0
    for imgFile in img_file_info:
        infiles.__dict__[imgFile.name] = imgFile.file_name
        numClassVars = numClassVars + len(imgFile.bands)

    outfiles = applier.FilenameAssociations()
    outfiles.outimage = out_prob_img
    otherargs = applier.OtherInputs()
    otherargs.classifier = classifier
    otherargs.mskVal = img_mask_val
    otherargs.numClassVars = numClassVars
    otherargs.imgFileInfo = img_file_info

    try:
        import tqdm

        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    print("Applying the Classifier")
    applier.apply(_applyXGBClassifier, infiles, outfiles, otherargs, controls=aControls)
    print("Completed")
    rsgislib.imageutils.pop_img_stats(
        out_prob_img, use_no_data=True, no_data_val=0, calc_pyramids=True
    )

    if out_class_img is not None:
        rsgislib.imagecalc.image_math(
            out_prob_img,
            out_class_img,
            "b1>{}?1:0".format(class_thres),
            gdalformat,
            rsgislib.TYPE_8UINT,
        )
        if gdalformat == "KEA":
            rsgislib.rastergis.pop_rat_img_stats(
                out_class_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
            )


def optimise_xgboost_multiclass_classifier(
    out_params_file,
    cls_info_dict,
    n_threads=1,
    mdl_cls_obj=None,
    sub_train_smpls=None,
    rnd_seed=42,
):
    """
    A function which performs a bayesian optimisation of the hyper-parameters for a multiclass xgboost
    classifier. A dict of class information, as ClassInfoObj objects, is defined with the training and
    validation data. Note, the training data inputted into this function might well be a smaller subset
    of the whole training dataset to speed up processing.

    This function requires that xgboost and skopt modules to be installed.

    :param out_params_file: The output model parameters which have been optimised.
    :param cls_info_dict: dict (key is string with class name) of ClassInfoObj objects defining the
                        training and validation data.
    :param n_threads: The number of threads to use to train the classifier.
    :param sub_train_smpls: Subset the training, if None or 0 then no sub-setting will occur. If
                            between 0-1 then a ratio subset (e.g., 0.25 = 25 % subset) will be taken.
                            If > 1 then that number of points will be taken per class.
    :param rnd_seed: the seed for the random selection of the training data.

    """
    if not HAVE_XGBOOST:
        raise rsgislib.RSGISPyException("Do not have xgboost module installed.")

    from skopt.space import Real, Integer
    from skopt import gp_minimize

    rnd_obj = numpy.random.RandomState(rnd_seed)

    n_classes = len(cls_info_dict)
    for clsname in cls_info_dict:
        if cls_info_dict[clsname].id >= n_classes:
            raise rsgislib.RSGISPyException(
                "ClassInfoObj '{}' id ({}) is not consecutive "
                "starting from 0.".format(clsname, cls_info_dict[clsname].id)
            )

    cls_data_dict = {}
    train_data_lst = []
    train_lbls_lst = []
    valid_data_lst = []
    valid_lbls_lst = []
    cls_ids = []
    n_classes = 0
    for clsname in cls_info_dict:
        sgl_cls_info = {}
        print("Reading Class {} Training".format(clsname))
        f = h5py.File(cls_info_dict[clsname].train_file_h5, "r")
        sgl_cls_info["train_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["train_data"] = numpy.array(f["DATA/DATA"])

        if (sub_train_smpls is not None) and (sub_train_smpls > 0):
            if sub_train_smpls < 1:
                sub_n_rows = int(sgl_cls_info["train_n_rows"] * sub_train_smpls)
            else:
                sub_n_rows = sub_train_smpls
            print("sub_n_rows = {}".format(sub_n_rows))
            if sub_n_rows > 0:
                sub_sel_rows = rnd_obj.choice(sgl_cls_info["train_n_rows"], sub_n_rows)
                sgl_cls_info["train_data"] = sgl_cls_info["train_data"][sub_sel_rows]
                sgl_cls_info["train_n_rows"] = sub_n_rows

        sgl_cls_info["train_data_lbls"] = numpy.zeros(
            sgl_cls_info["train_n_rows"], dtype=int
        )
        sgl_cls_info["train_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()

        train_data_lst.append(sgl_cls_info["train_data"])
        train_lbls_lst.append(sgl_cls_info["train_data_lbls"])

        print("Reading Class {} Validation".format(clsname))
        f = h5py.File(cls_info_dict[clsname].valid_file_h5, "r")
        sgl_cls_info["valid_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["valid_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["valid_data_lbls"] = numpy.zeros(
            sgl_cls_info["valid_n_rows"], dtype=int
        )
        sgl_cls_info["valid_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        valid_data_lst.append(sgl_cls_info["valid_data"])
        valid_lbls_lst.append(sgl_cls_info["valid_data_lbls"])

        cls_data_dict[clsname] = sgl_cls_info
        cls_ids.append(cls_info_dict[clsname].id)
        n_classes = n_classes + 1

    print("Finished Reading Data")

    vaild_np = numpy.concatenate(valid_data_lst)
    vaild_lbl_np = numpy.concatenate(valid_lbls_lst)
    d_valid = xgb.DMatrix(vaild_np, label=vaild_lbl_np)

    d_train = xgb.DMatrix(
        numpy.concatenate(train_data_lst), label=numpy.concatenate(train_lbls_lst)
    )

    space = [
        Real(0.01, 0.9, name="eta"),
        Integer(0, 100, name="gamma"),
        Integer(2, 20, name="max_depth"),
        Integer(1, 10, name="min_child_weight"),
        Integer(0, 10, name="max_delta_step"),
        Real(0.5, 1, name="subsample"),
        Integer(2, 100, name="num_boost_round"),
    ]

    def _objective(values):
        params = {
            "eta": values[0],
            "gamma": values[1],
            "max_depth": values[2],
            "min_child_weight": values[3],
            "max_delta_step": values[4],
            "subsample": values[5],
            "nthread": n_threads,
            "eval_metric": "merror",
            "objective": "multi:softmax",
            "num_class": n_classes,
        }

        print("\nNext set of params.....", params)

        num_boost_round = values[6]
        print("num_boost_round = {}.".format(num_boost_round))

        watchlist = [(d_train, "train"), (d_valid, "validation")]
        evals_results = {}
        model_xgb = xgb.train(
            params,
            d_train,
            num_boost_round,
            evals=watchlist,
            evals_result=evals_results,
            verbose_eval=False,
            xgb_model=mdl_cls_obj,
        )

        vld_preds_idxs = model_xgb.predict(d_valid)

        acc_score = -accuracy_score(vaild_lbl_np, vld_preds_idxs)
        print("\nAccScore.....", -acc_score, ".....iter.....")
        gc.collect()
        return acc_score

    res_gp = gp_minimize(
        _objective, space, n_calls=20, random_state=0, n_random_starts=10
    )

    print("Best score={}".format(res_gp.fun))
    best_params = res_gp.x
    print("Best Params:\n{}".format(best_params))

    params = {
        "eta": float(best_params[0]),
        "gamma": int(best_params[1]),
        "max_depth": int(best_params[2]),
        "min_child_weight": int(best_params[3]),
        "max_delta_step": int(best_params[4]),
        "subsample": float(best_params[5]),
        "nthread": int(n_threads),
        "eval_metric": "merror",
        "objective": "multi:softmax",
        "num_class": int(n_classes),
        "num_boost_round": int(best_params[6]),
    }

    with open(out_params_file, "w") as fp:
        json.dump(
            params,
            fp,
            sort_keys=True,
            indent=4,
            separators=(",", ": "),
            ensure_ascii=False,
        )


def train_xgboost_multiclass_classifier(
    out_mdl_file, cls_params_file, cls_info_dict, n_threads=1, mdl_cls_obj=None
):
    """
    A function which performs a bayesian optimisation of the hyper-parameters for a multiclass xgboost
    classifier producing a full trained model at the end. A dict of class information, as ClassInfoObj
    objects, is defined with the training data.

    This function requires that xgboost modules to be installed.

    :param out_mdl_file: The output model which can be loaded to perform a classification.
    :param cls_params_file: A JSON file with the model parameters
    :param cls_info_dict: dict (key is string with class name) of ClassInfoObj objects defining the training data.
    :param n_threads: The number of threads to use to train the classifier.

    """
    if not HAVE_XGBOOST:
        raise rsgislib.RSGISPyException("Do not have xgboost module installed.")

    n_classes = len(cls_info_dict)
    for clsname in cls_info_dict:
        if cls_info_dict[clsname].id >= n_classes:
            raise rsgislib.RSGISPyException(
                "ClassInfoObj '{}' id ({}) is not consecutive "
                "starting from 0.".format(clsname, cls_info_dict[clsname].id)
            )

    cls_data_dict = {}
    train_data_lst = []
    train_lbls_lst = []
    valid_data_lst = []
    valid_lbls_lst = []
    test_data_lst = []
    test_lbls_lst = []
    cls_ids = []
    n_classes = 0
    for clsname in cls_info_dict:
        sgl_cls_info = {}
        print("Reading Class {} Training".format(clsname))
        f = h5py.File(cls_info_dict[clsname].train_file_h5, "r")
        sgl_cls_info["train_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["train_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["train_data_lbls"] = numpy.zeros(
            sgl_cls_info["train_n_rows"], dtype=int
        )
        sgl_cls_info["train_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        train_data_lst.append(sgl_cls_info["train_data"])
        train_lbls_lst.append(sgl_cls_info["train_data_lbls"])

        print("Reading Class {} Validation".format(clsname))
        f = h5py.File(cls_info_dict[clsname].valid_file_h5, "r")
        sgl_cls_info["valid_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["valid_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["valid_data_lbls"] = numpy.zeros(
            sgl_cls_info["valid_n_rows"], dtype=int
        )
        sgl_cls_info["valid_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        valid_data_lst.append(sgl_cls_info["valid_data"])
        valid_lbls_lst.append(sgl_cls_info["valid_data_lbls"])

        print("Reading Class {} Testing".format(clsname))
        f = h5py.File(cls_info_dict[clsname].test_file_h5, "r")
        sgl_cls_info["test_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["test_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["test_data_lbls"] = numpy.zeros(
            sgl_cls_info["test_n_rows"], dtype=int
        )
        sgl_cls_info["test_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        test_data_lst.append(sgl_cls_info["test_data"])
        test_lbls_lst.append(sgl_cls_info["test_data_lbls"])

        cls_data_dict[clsname] = sgl_cls_info
        cls_ids.append(cls_info_dict[clsname].id)
        n_classes = n_classes + 1

    print("Finished Reading Data")

    vaild_np = numpy.concatenate(valid_data_lst)
    vaild_lbl_np = numpy.concatenate(valid_lbls_lst)
    d_valid = xgb.DMatrix(vaild_np, label=vaild_lbl_np)

    d_train = xgb.DMatrix(
        numpy.concatenate(train_data_lst), label=numpy.concatenate(train_lbls_lst)
    )

    test_np = numpy.concatenate(test_data_lst)
    test_lbl_np = numpy.concatenate(test_lbls_lst)
    d_test = xgb.DMatrix(test_np, label=test_lbl_np)

    with open(cls_params_file) as f:
        cls_params = json.load(f)

    if n_classes != cls_params["num_class"]:
        raise rsgislib.RSGISPyException(
            "The number of classes used to optimise the "
            "classifier and the number for training are different!"
        )

    print("Start Training Find Classifier")

    params = {
        "eta": cls_params["eta"],
        "gamma": cls_params["gamma"],
        "max_depth": cls_params["max_depth"],
        "min_child_weight": cls_params["min_child_weight"],
        "max_delta_step": cls_params["max_delta_step"],
        "subsample": cls_params["subsample"],
        "nthread": n_threads,
        "eval_metric": cls_params["eval_metric"],
        "objective": cls_params["objective"],
        "num_class": n_classes,
    }

    num_boost_round = cls_params["num_boost_round"]

    watchlist = [(d_train, "train"), (d_valid, "validation")]
    evals_results = {}
    model_xgb = xgb.train(
        params,
        d_train,
        num_boost_round,
        evals=watchlist,
        evals_result=evals_results,
        verbose_eval=False,
        xgb_model=mdl_cls_obj,
    )
    model_xgb.save_model(out_mdl_file)

    vld_preds_idxs = model_xgb.predict(d_valid)
    valid_acc_scr = accuracy_score(vaild_lbl_np, vld_preds_idxs)
    print("Validate Accuracy: {}".format(valid_acc_scr))

    test_preds_idxs = model_xgb.predict(d_test)
    test_acc_scr = accuracy_score(test_lbl_np, test_preds_idxs)
    print("Testing Accuracy: {}".format(test_acc_scr))


def train_opt_xgboost_multiclass_classifier(
    out_mdl_file, cls_info_dict, n_threads=1, mdl_cls_obj=None
):
    """
    A function which performs a bayesian optimisation of the hyper-parameters for a multiclass xgboost
    classifier producing a full trained model at the end. A dict of class information, as ClassInfoObj
    objects, is defined with the training data.

    This function requires that xgboost and skopt modules to be installed.

    :param out_mdl_file: The output model which can be loaded to perform a classification.
    :param cls_info_dict: dict (key is string with class name) of ClassInfoObj objects defining the training data.
    :param n_threads: The number of threads to use to train the classifier.

    """
    if not HAVE_XGBOOST:
        raise rsgislib.RSGISPyException("Do not have xgboost module installed.")

    from skopt.space import Real, Integer
    from skopt import gp_minimize

    n_classes = len(cls_info_dict)
    for clsname in cls_info_dict:
        if cls_info_dict[clsname].id >= n_classes:
            raise rsgislib.RSGISPyException(
                "ClassInfoObj '{}' id ({}) is not consecutive starting from 0.".format(
                    clsname, cls_info_dict[clsname].id
                )
            )

    cls_data_dict = {}
    train_data_lst = []
    train_lbls_lst = []
    valid_data_lst = []
    valid_lbls_lst = []
    test_data_lst = []
    test_lbls_lst = []
    cls_ids = []
    n_classes = 0
    for clsname in cls_info_dict:
        sgl_cls_info = {}
        print("Reading Class {} Training".format(clsname))
        f = h5py.File(cls_info_dict[clsname].train_file_h5, "r")
        sgl_cls_info["train_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["train_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["train_data_lbls"] = numpy.zeros(
            sgl_cls_info["train_n_rows"], dtype=int
        )
        sgl_cls_info["train_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        train_data_lst.append(sgl_cls_info["train_data"])
        train_lbls_lst.append(sgl_cls_info["train_data_lbls"])

        print("Reading Class {} Validation".format(clsname))
        f = h5py.File(cls_info_dict[clsname].valid_file_h5, "r")
        sgl_cls_info["valid_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["valid_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["valid_data_lbls"] = numpy.zeros(
            sgl_cls_info["valid_n_rows"], dtype=int
        )
        sgl_cls_info["valid_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        valid_data_lst.append(sgl_cls_info["valid_data"])
        valid_lbls_lst.append(sgl_cls_info["valid_data_lbls"])

        print("Reading Class {} Testing".format(clsname))
        f = h5py.File(cls_info_dict[clsname].test_file_h5, "r")
        sgl_cls_info["test_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["test_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["test_data_lbls"] = numpy.zeros(
            sgl_cls_info["test_n_rows"], dtype=int
        )
        sgl_cls_info["test_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        test_data_lst.append(sgl_cls_info["test_data"])
        test_lbls_lst.append(sgl_cls_info["test_data_lbls"])

        cls_data_dict[clsname] = sgl_cls_info
        cls_ids.append(cls_info_dict[clsname].id)
        n_classes = n_classes + 1

    print("Finished Reading Data")

    vaild_np = numpy.concatenate(valid_data_lst)
    vaild_lbl_np = numpy.concatenate(valid_lbls_lst)
    d_valid = xgb.DMatrix(vaild_np, label=vaild_lbl_np)

    d_train = xgb.DMatrix(
        numpy.concatenate(train_data_lst), label=numpy.concatenate(train_lbls_lst)
    )

    test_np = numpy.concatenate(test_data_lst)
    test_lbl_np = numpy.concatenate(test_lbls_lst)
    d_test = xgb.DMatrix(test_np, label=test_lbl_np)

    space = [
        Real(0.01, 0.9, name="eta"),
        Integer(0, 100, name="gamma"),
        Integer(2, 20, name="max_depth"),
        Integer(1, 10, name="min_child_weight"),
        Integer(0, 10, name="max_delta_step"),
        Real(0.5, 1, name="subsample"),
        Integer(2, 100, name="num_boost_round"),
    ]

    def _objective(values):
        params = {
            "eta": values[0],
            "gamma": values[1],
            "max_depth": values[2],
            "min_child_weight": values[3],
            "max_delta_step": values[4],
            "subsample": values[5],
            "nthread": n_threads,
            "eval_metric": "merror",
            "objective": "multi:softmax",
            "num_class": n_classes,
        }

        print("\nNext set of params.....", params)

        num_boost_round = values[6]
        print("num_boost_round = {}.".format(num_boost_round))

        watchlist = [(d_train, "train"), (d_valid, "validation")]
        evals_results = {}
        model_xgb = xgb.train(
            params,
            d_train,
            num_boost_round,
            evals=watchlist,
            evals_result=evals_results,
            verbose_eval=False,
            xgb_model=mdl_cls_obj,
        )

        vld_preds_idxs = model_xgb.predict(d_valid)

        acc_score = -accuracy_score(vaild_lbl_np, vld_preds_idxs)
        print("\nAccScore.....", -acc_score, ".....iter.....")
        gc.collect()
        return acc_score

    res_gp = gp_minimize(
        _objective, space, n_calls=20, random_state=0, n_random_starts=10
    )

    print("Best score={}".format(res_gp.fun))
    best_params = res_gp.x
    print("Best Params:\n{}".format(best_params))

    print("Start Training Find Classifier")
    params = {
        "eta": best_params[0],
        "gamma": best_params[1],
        "max_depth": best_params[2],
        "min_child_weight": best_params[3],
        "max_delta_step": best_params[4],
        "subsample": best_params[5],
        "nthread": n_threads,
        "eval_metric": "merror",
        "objective": "multi:softmax",
        "num_class": n_classes,
    }

    num_boost_round = best_params[6]

    watchlist = [(d_train, "train"), (d_valid, "validation")]
    evals_results = {}
    model_xgb = xgb.train(
        params,
        d_train,
        num_boost_round,
        evals=watchlist,
        evals_result=evals_results,
        verbose_eval=False,
        xgb_model=mdl_cls_obj,
    )
    model_xgb.save_model(out_mdl_file)

    vld_preds_idxs = model_xgb.predict(d_valid)
    valid_acc_scr = accuracy_score(vaild_lbl_np, vld_preds_idxs)
    print("Validate Accuracy: {}".format(valid_acc_scr))

    test_preds_idxs = model_xgb.predict(d_test)
    test_acc_scr = accuracy_score(test_lbl_np, test_preds_idxs)
    print("Testing Accuracy: {}".format(test_acc_scr))


def apply_xgboost_multiclass_classifier(
    class_train_info,
    model_file,
    in_mask_img,
    img_mask_val,
    img_file_info,
    out_class_img,
    gdalformat,
    class_clr_names=True,
    n_threads=1,
):
    """
    This function applies a trained multiple classes xgboost model. The function train_xgboost_multiclass_classifier
    can be used to train such as model. The output image will contain the probability of membership to the class of
    interest. You will need to threshold this image to get a final hard classification. Alternative, a hard class
    output image and threshold can be applied to this image.

    :param class_train_info: dict (where the key is the class name) of rsgislib.classification.ClassInfoObj
                           objects which will be used to train the classifier (i.e., train_xgboost_multiclass_classifier()),
                           provide pixel value id and RGB class values.
    :param model_file: a trained xgboost multiclass model which can be loaded with lgb.Booster(model_file=model_file).
    :param in_mask_img: is an image file providing a mask to specify where should be classified. Simplest mask is all the
                    valid data regions (rsgislib.imageutils.gen_valid_mask)
    :param img_mask_val: the pixel value within the imgMask to limit the region to which the classification is applied.
                       Can be used to create a heirachical classification.
    :param img_file_info: a list of rsgislib.imageutils.ImageBandInfo objects (also used within
                        rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) to identify which images and bands are to
                        be used for the classification so it adheres to the training data.
    :param out_class_img: Output image which will contain the hard classification defined as the maximum probability.
    :param gdalformat: is the output image format - all GDAL supported formats are supported.
    :param class_clr_names: default is True and therefore a colour table will the colours specified in ClassInfoObj
                          and a class_names (from classTrainInfo) column will be added to the output file.
    :param n_threads: The number of threads to use for the classifier.

    """
    if not HAVE_XGBOOST:
        raise rsgislib.RSGISPyException("Do not have xgboost module installed.")

    def _applyXGBMClassifier(info, inputs, outputs, otherargs):
        outClassIdVals = numpy.zeros_like(inputs.imageMask, dtype=numpy.uint16)
        if numpy.any(inputs.imageMask == otherargs.mskVal):
            n_pxls = inputs.imageMask.shape[1] * inputs.imageMask.shape[2]
            outClassIdVals = outClassIdVals.flatten()
            imgMaskVals = inputs.imageMask.flatten()
            classVars = numpy.zeros(
                (n_pxls, otherargs.numClassVars), dtype=numpy.float32
            )
            # Array index which can be used to populate the output array following masking etc.
            ID = numpy.arange(imgMaskVals.shape[0])
            classVarsIdx = 0
            for imgFile in otherargs.imgFileInfo:
                imgArr = inputs.__dict__[imgFile.name]
                for band in imgFile.bands:
                    classVars[..., classVarsIdx] = imgArr[(band - 1)].flatten()
                    classVarsIdx = classVarsIdx + 1
            classVars = classVars[imgMaskVals == otherargs.mskVal]
            ID = ID[imgMaskVals == otherargs.mskVal]
            preds_idxs = otherargs.classifier.predict(xgb.DMatrix(classVars))
            preds_cls_ids = numpy.zeros_like(preds_idxs, dtype=numpy.uint16)
            for cld_id, idx in zip(
                otherargs.cls_id_lut, numpy.arange(0, len(otherargs.cls_id_lut))
            ):
                preds_cls_ids[preds_idxs == idx] = cld_id

            outClassIdVals[ID] = preds_cls_ids
            outClassIdVals = numpy.expand_dims(
                outClassIdVals.reshape(
                    (inputs.imageMask.shape[1], inputs.imageMask.shape[2])
                ),
                axis=0,
            )

        outputs.outclsimage = outClassIdVals

    classifier = xgb.Booster({"nthread": n_threads})
    classifier.load_model(model_file)

    infiles = applier.FilenameAssociations()
    infiles.imageMask = in_mask_img
    numClassVars = 0
    for imgFile in img_file_info:
        infiles.__dict__[imgFile.name] = imgFile.file_name
        numClassVars = numClassVars + len(imgFile.bands)

    n_classes = len(class_train_info)
    cls_id_lut = numpy.zeros(n_classes)
    for clsname in class_train_info:
        if class_train_info[clsname].id >= n_classes:
            raise rsgislib.RSGISPyException(
                "ClassInfoObj '{}' id ({}) is not consecutive starting from 0.".format(
                    clsname, class_train_info[clsname].id
                )
            )
        cls_id_lut[class_train_info[clsname].id] = class_train_info[clsname].out_id

    outfiles = applier.FilenameAssociations()
    outfiles.outclsimage = out_class_img
    otherargs = applier.OtherInputs()
    otherargs.classifier = classifier
    otherargs.mskVal = img_mask_val
    otherargs.numClassVars = numClassVars
    otherargs.imgFileInfo = img_file_info
    otherargs.n_classes = n_classes
    otherargs.cls_id_lut = cls_id_lut

    try:
        import tqdm

        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    print("Applying the Classifier")
    applier.apply(
        _applyXGBMClassifier, infiles, outfiles, otherargs, controls=aControls
    )
    print("Completed Classification")

    if class_clr_names:
        rsgislib.rastergis.pop_rat_img_stats(
            out_class_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
        )
        ratDataset = gdal.Open(out_class_img, gdal.GA_Update)
        red = rat.readColumn(ratDataset, "Red")
        green = rat.readColumn(ratDataset, "Green")
        blue = rat.readColumn(ratDataset, "Blue")
        class_names = numpy.empty_like(red, dtype=numpy.dtype("a255"))
        class_names[...] = ""

        for classKey in class_train_info:
            print("Apply Colour to class '" + classKey + "'")
            red[class_train_info[classKey].out_id] = class_train_info[classKey].red
            green[class_train_info[classKey].out_id] = class_train_info[classKey].green
            blue[class_train_info[classKey].out_id] = class_train_info[classKey].blue
            class_names[class_train_info[classKey].out_id] = classKey

        rat.writeColumn(ratDataset, "Red", red)
        rat.writeColumn(ratDataset, "Green", green)
        rat.writeColumn(ratDataset, "Blue", blue)
        rat.writeColumn(ratDataset, "class_names", class_names)
        ratDataset = None


def apply_xgboost_multiclass_classifier_rat(
    clumps_img,
    variables,
    model_file,
    class_train_info,
    out_col_int="OutClass",
    out_col_str="OutClassName",
    roi_col=None,
    roi_val=1,
    class_colours=True,
    nthread=1,
):
    """
    A function which will apply an XGBoost model within a Raster Attribute Table (RAT).

    :param clumps_img: is the clumps image on which the classification is to be performed
    :param variables: is an array of column names which are to be used for the classification
    :param class_train_info: dict (where the key is the class name) of
                             rsgislib.classification.ClassInfoObj objects which will be
                             used to train the classifier (i.e.,
                             train_xgboost_multiclass_classifier()), provide pixel value
                             id and RGB class values.
    :param model_file: a trained xgboost multiclass model which can be loaded with lgb.Booster(model_file=model_file).
    :param out_col_int: is the output column name for the int class representation (Default: 'OutClass')
    :param out_col_str: is the output column name for the class names column (Default: 'OutClassName')
    :param roi_col: is a column name for a column which specifies the region to be classified. If None ignored (Default: None)
    :param roi_val: is a int value used within the roi_col to select a region to be classified (Default: 1)
    :param class_colours: is a boolean specifying whether the RAT colour table should be
                          updated using the classification colours (default: True)
    :param nthread: The number of threads to use for the classifier."""
    if not HAVE_XGBOOST:
        raise rsgislib.RSGISPyException("Do not have xgboost module installed.")

    from rios import ratapplier

    def _apply_rat_classifier(info, inputs, outputs, otherargs):
        """
        This function is used internally within classify_within_rat_tiled using the RIOS ratapplier function
        """
        numpyVars = []
        for var in otherargs.vars:
            varVals = getattr(inputs.inrat, var)
            numpyVars.append(varVals)

        xData = numpy.array(numpyVars)
        xData = xData.transpose()

        ID = numpy.arange(xData.shape[0])
        outClassIntVals = numpy.zeros(xData.shape[0], dtype=numpy.int16)
        outClassNamesVals = numpy.empty(xData.shape[0], dtype=numpy.dtype("a255"))
        outClassNamesVals[...] = ""

        ID = ID[numpy.isfinite(xData).all(axis=1)]
        vData = xData[numpy.isfinite(xData).all(axis=1)]

        if otherargs.roiCol is not None:
            roi = getattr(inputs.inrat, otherargs.roiCol)
            roi = roi[numpy.isfinite(xData).all(axis=1)]
            vData = vData[roi == otherargs.roiVal]
            ID = ID[roi == otherargs.roiVal]

        preds_idxs = otherargs.classifier.predict(xgb.DMatrix(vData))

        preds_cls_ids = numpy.zeros_like(preds_idxs, dtype=numpy.uint16)
        for cld_id, idx in zip(
            otherargs.cls_id_lut, numpy.arange(0, len(otherargs.cls_id_lut))
        ):
            preds_cls_ids[preds_idxs == idx] = cld_id

        outClassIntVals[ID] = preds_cls_ids
        setattr(outputs.outrat, otherargs.outColInt, outClassIntVals)

        for cls_id in otherargs.cls_name_lut:
            outClassNamesVals[outClassIntVals == cls_id] = otherargs.cls_name_lut[
                cls_id
            ]
        setattr(outputs.outrat, otherargs.outColStr, outClassNamesVals)

        if otherargs.class_colours:
            red = getattr(inputs.inrat, "Red")
            green = getattr(inputs.inrat, "Green")
            blue = getattr(inputs.inrat, "Blue")

            # Set Background to black
            red[...] = 0
            green[...] = 0
            blue[...] = 0

            # Set colours
            for class_name in otherargs.class_train_info:
                cls_id = otherargs.class_train_info[class_name].out_id
                red = numpy.where(
                    outClassIntVals == cls_id,
                    otherargs.class_train_info[class_name].red,
                    red,
                )
                green = numpy.where(
                    outClassIntVals == cls_id,
                    otherargs.class_train_info[class_name].green,
                    green,
                )
                blue = numpy.where(
                    outClassIntVals == cls_id,
                    otherargs.class_train_info[class_name].blue,
                    blue,
                )

            setattr(outputs.outrat, "Red", red)
            setattr(outputs.outrat, "Green", green)
            setattr(outputs.outrat, "Blue", blue)

    classifier = xgb.Booster({"nthread": nthread})
    classifier.load_model(model_file)

    n_classes = len(class_train_info)
    cls_id_lut = numpy.zeros(n_classes)
    cls_name_lut = dict()
    for clsname in class_train_info:
        if class_train_info[clsname].id >= n_classes:
            raise rsgislib.RSGISPyException(
                "ClassInfoObj '{}' id ({}) is not consecutive starting from 0.".format(
                    clsname, class_train_info[clsname].id
                )
            )
        cls_id_lut[class_train_info[clsname].id] = class_train_info[clsname].out_id
        cls_name_lut[class_train_info[clsname].out_id] = clsname

    in_rats = ratapplier.RatAssociations()
    out_rats = ratapplier.RatAssociations()
    in_rats.inrat = ratapplier.RatHandle(clumps_img)
    out_rats.outrat = ratapplier.RatHandle(clumps_img)

    otherargs = ratapplier.OtherArguments()
    otherargs.vars = variables
    otherargs.classifier = classifier
    otherargs.outColInt = out_col_int
    otherargs.outColStr = out_col_str
    otherargs.roiCol = roi_col
    otherargs.roiVal = roi_val
    otherargs.n_classes = n_classes
    otherargs.cls_id_lut = cls_id_lut
    otherargs.cls_name_lut = cls_name_lut
    otherargs.class_colours = class_colours
    otherargs.class_train_info = class_train_info

    try:
        import tqdm

        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    aControls = applier.ApplierControls()
    aControls.progress = progress_bar

    ratapplier.apply(
        _apply_rat_classifier, in_rats, out_rats, otherargs=otherargs, controls=None
    )
